"""
Utilities for splitting a protein dataset based on sequence homology.

NB. These functions require Blast+ for proper execution. On Linux, this can be installed with:

sudo apt install ncbi-blast+

Otherwise, please see: https://blast.ncbi.nlm.nih.gov/Blast.cgi?CMD=Web&PAGE_TYPE=BlastDocs&DOC_TYPE=Download
"""

import contextlib
import os
import random
import sys
from typing import Dict, List, Optional

import networkx as nx
import numpy as np
from Bio import SeqIO


def build_fasta_file_from_mapping(
    pdb_sequence_mapping: Dict[str, str],
    fasta_out: str,
):
    """
    Builds a fasta database from a mapping of PDBs (or other identifier) to sequences.

    :param pdb_sequence_mapping: dictionary with the sequence of each pdb file
    :type pdb_sequence_mapping: Dict[str, str]
    :param fasta_out: file name of the fasta database
    :type fasta_out: str
    """
    with open(fasta_out, mode="w") as f_out:
        for k, v in pdb_sequence_mapping.items():
            seq_id = k
            print(f">{seq_id}\n{v}", file=f_out)


def build_fasta_file_from_graphs(
    graphs: List[nx.Graph], fasta_out: str, chains: Optional[List[str]] = None
):
    if chains is None:
        chains = ["A"] * len(graphs)
    mapping = {
        f"{g.name}_{chain}": g.graph[f"sequence_{chain}"]
        for g, chain in zip(graphs, chains)
    }

    build_fasta_file_from_mapping(mapping, fasta_out)


def get_seq_records(
    filename,
    alphabet=None,  # alphabet=generic_protein,
    check_sequences=False,
    file_format="fasta",
    return_as_dictionary=False,
):
    """
    Get the sequence records from a file.

    :param filename: File with sequences [fasta in our case].
    :type filename: str
    :param alphabet: alphabet to be used.
    :param check_sequences: Sanity check on the sequences.
    :type check_sequences: bool
    :param file_format: We use fasta format (don't change this).
    :type file_format: str
    :param return_as_dictionary: if ``True`` uses ``records.to_dict()`` to convert the record into a dictionary and return it.
    :type return_as_dictionary: bool
    :return: a list of the ``SeqRecord`` of the sequences.
             - ``.seq``: access the records (sequence, i.e. ``Seq``, for actual string use ``str(records[].seq))``
             - ``.id``: access the ids in the file
             - ``.name``: access names (sometimes equal to ``id``, depends on format)
             - ``.description``: is the description (when included)
             - ``.dbxref`` which (if present) is a list with the cross-referenced database
    Notes:
    ------
    For a list of available file_format see http://biopython.org/wiki/SeqIO#File_Formats .aln is generally 'clustal'
    """
    if alphabet is None and check_sequences:
        sys.stderr.write(
            f"WARNING in {get_seq_records.__name__} ask to check_sequences but no "
            "Alphabet given. Only checking for terminating *!\n"
        )
        check_sequences = False
    with open(filename, "r") as handle:
        records = list(SeqIO.parse(handle, file_format, alphabet=alphabet))
    del handle
    if check_sequences:
        for record in records:
            if record.seq[-1] == "*":
                record.seq = record.seq[:-1]
                if alphabet is not None:
                    good = False
                    with contextlib.suppress(ValueError):
                        good = _verify_alphabet(record.seq)
                    if not good:
                        sys.stderr.write(
                            f"WARNING in {get_seq_records.__name__} sequence {record.seq.id} from file "
                            f"{filename} is not compatible with declared alphabet {str(alphabet)}\n"
                        )
    return SeqIO.to_dict(records) if return_as_dictionary else records


def create_pairs_for_clustering(
    fasta_file: str,
    seq_id_low_thresh: float = 25.0,
    use_very_loose_condition: bool = False,
    out_file: Optional[str] = None,
    n_cpu: int = 1,
    max_target_seqs: int = 200,
    delete_blast_databases_when_done: bool = False,
):
    """
    Create sequences pairs for clustering [sequence names MUST be the ids!].

    :param fasta_file: file with fastas
    :type fasta_file: str
    :param seq_id_low_thresh: sequence identity of lower threshold
    :type seq_id_low_thresh: float
    :param use_very_loose_condition: ??
    :type use_very_loose_condition: bool
    :param out_file: name of the output file
    :type out_file: str
    :param n_cpu: number of cpus to be used
    :type n_cpu: int
    :param max_target_seqs: ??
    :type max_target_seqs: int
    :param delete_blast_databases_when_done: delete the db once the process is finished
    :type delete_blast_databases_when_done: bool
    :return:
    """
    if out_file is None:
        out_file = "pairs_for_clustering.txt"
    rec = get_seq_records(fasta_file)
    # blast each sequence on the database to get the identity pairs
    pairs = []
    corresponding_identities = []
    for s in rec:
        pairs += [(s.id, s.id)]
        corresponding_identities += [100.0]
        b_res = blast(
            s,
            fasta_file,
            max_target_seqs=max_target_seqs,
            n_cpu=n_cpu,
            delete_blast_file=True,
        )
        if (
            b_res is None
        ):  # blast sometimes does not hit the very same sequence (with blst+ it should)
            continue
        if use_very_loose_condition:
            _, d_res = process_blast_results(
                b_res,
                len(s.seq),
                return_dictionary_of_numpy_matrices=True,
                filter_best_match_for_same_subject=False,
                add_identity_on_shortest=False,
            )
            for k in d_res:
                if k == s.id:
                    continue  # we added this manually at the beginning
                max_id = max(
                    d_res[k][:, 0]
                )  # maximum identity between different possible alignments
                if not 0 <= max_id <= 100:
                    print(f"Max_id problem: {max_id}, {k}, {s.id}")
                if (
                    max_id > seq_id_low_thresh
                    and frozenset([k, s.id]) not in pairs
                ):
                    pairs += [frozenset([s.id, k])]
                    corresponding_identities += [max_id]
        else:
            b_res = process_blast_results(
                b_res,
                len(s.seq),
                return_dictionary_of_numpy_matrices=False,
                filter_best_match_for_same_subject=True,
                add_identity_on_shortest=True,
            )
            for p in b_res:
                if p[0] == s.id:
                    continue  # we added this manually at the beginning
                if p[-1] > seq_id_low_thresh:
                    pairs += [frozenset([s.id, p[0]])]
                    corresponding_identities += [p[-1]]

    if out_file is not None:
        out = open(out_file, "w")
    done = {}
    for j, p in enumerate(pairs):
        if out_file is not None:
            if p in done:
                continue
            ff = list(p)
            out.write(
                "%s\t%s\t%lf\n" % (ff[0], ff[1], corresponding_identities[j])
            )
        done[p] = corresponding_identities[j]
    if out_file is not None:
        out.close()
    if delete_blast_databases_when_done:
        os.system(f"rm -f {fasta_file}.phr {fasta_file}.pin {fasta_file}.psq")
    return done


def clustering_from_pairs(pairs, out_file=None):
    """
    Clustering of a set of elements given a list of 'homologous' with possibly false negatives
    :param pairs: pair of sequences
    :param out_file: output file
    :return: the clusters
    """
    clusters = []
    j = 0
    key_to_cluster = {}
    for p in pairs:
        p = list(p)
        if p[0] in key_to_cluster and p[1] not in key_to_cluster:
            if p[1] not in clusters[key_to_cluster[p[0]]]:
                clusters[key_to_cluster[p[0]]].append(p[1])
            key_to_cluster[p[1]] = key_to_cluster[p[0]]
        elif p[1] in key_to_cluster and p[0] not in key_to_cluster:
            if p[0] not in clusters[key_to_cluster[p[1]]]:
                clusters[key_to_cluster[p[1]]].append(p[0])
            key_to_cluster[p[0]] = key_to_cluster[p[1]]
        elif p[1] in key_to_cluster:
            if key_to_cluster[p[1]] != key_to_cluster[p[0]]:
                # we leave p1 empty and we put all in p0, loop backward to remove stuff...
                del_id = key_to_cluster[p[1]]
                c = len(clusters[del_id]) - 1
                while c >= 0:
                    el = clusters[del_id][c]
                    if el in key_to_cluster:
                        del key_to_cluster[el]
                    if el not in clusters[key_to_cluster[p[0]]]:
                        clusters[key_to_cluster[p[0]]].append(el)
                        key_to_cluster[el] = key_to_cluster[p[0]]
                    del clusters[del_id][c]
                    c -= 1
        else:  # add a new cluster
            key_to_cluster[p[0]] = j
            if p[0] == p[1]:
                clusters.append([p[0]])
            else:
                clusters.append([p[0], p[1]])
                key_to_cluster[p[1]] = j
            j += 1
    del key_to_cluster
    c = len(clusters) - 1
    while c >= 0:
        if not clusters[c]:
            del clusters[c]
        c -= 1
    if out_file is not None:
        with open(out_file, "w") as out:
            for c in clusters:
                out.write("%s\n" % (" ".join(c)))
    return clusters


def blast(
    sequence,
    database,
    ungapped=False,
    add_sseq=False,
    add_qseq=False,
    tmp_folder="/tmp/",
    blast_path="",
    sequence_name="",
    blast_filename="blast" + str(os.getpid()) + ".txt",
    max_target_seqs=100,
    n_cpu=1,
    delete_blast_file=True,
):
    """
    Given a sequence and a database file (e.g. a .fasta file with some sequences) blasts the sequence on the database

    :param sequence: sequence of reference
    :param database: database against which to blast the sequence
    :param return_csv_data_class: Returns a dict with as keys subject_id and as value the list (without subject_id).
        This also contain a header under .hd
    :param ungapped:
    :param add_sseq: append to the returned list the keywords sseq (aligned part of subject sequence)
    :param add_qseq: append to the returned list the keywords qseq (aligned part of query sequence)
    :param tmp_folder: temporary folder
    :param blast_path: path to blast installation folder
    :param sequence_name: name of the sequence
    :param blast_filename: filename for the blast file
    :param max_target_seqs: maximum number of target sequences
    :param n_cpu: number of cpus
    :param delete_blast_file: if True deletes the blast file at the end
    :return: a list of lists.
             - each list is a possible match (best is at index 0)
             - if there is no matching it returns None
             - The inner part is:
             [subject_id (name of the match), identity (percentage), alignment_length,
             mismatches (number within alignment length), gaps (number within alignment length),
             q_start (start index in the query sequence), q_end (end index),
             s_start (start index in the matching sequence), s_end (end index), evalue,
             nident (number of identical amino acids in aligned region), slen (sublect length)]
    Notes
    -----
    Check that NCBI blast is correctly installed in the folder blast_path
    """

    if tmp_folder is None:
        tmp_folder = ""
    elif tmp_folder != "":
        if tmp_folder[-1] != "/":
            tmp_folder += "/"
        if delete_blast_file and tmp_folder not in blast_filename:
            blast_filename = tmp_folder + blast_filename
    if type(sequence) is not str:  # we assume is a Seq object
        sequence_name = sequence.id
        sequence = str(sequence.seq)
    if blast_path != "" and not os.path.isfile(
        blast_path + "makeblastdb"
    ):  # check if the BLAST_PATH is correct
        raise IOError(
            f"***ERROR*** in run_blast path {blast_path} doesnt lead to blast directory "
            "where makeblastdb should be located"
        )
    seq_file = open(tmp_folder + "sequence_temp.txt", "w")
    seq_file.write(">" + sequence_name + "\n" + sequence + "\n")
    seq_file.close()

    # check if the blast database has already been built, if not build it
    if not (
        os.path.isfile(database + ".phr")
        and os.path.isfile(database + ".pin")
        and os.path.isfile(database + ".psq")
    ):
        try:
            os.system(blast_path + "makeblastdb -dbtype prot -in " + database)
        except Exception:
            print(
                "***ERROR*** in run_blast() cannot build blast database maybe you wish to "
                "set BLAST_PATH to correct directory"
            )
            raise
    ungapped = " -comp_based_stats F  -ungapped" if ungapped else ""
    table_k = "6 sseqid pident length mismatch gapopen qstart qend sstart send evalue nident slen"
    if add_sseq:
        table_k += " sseq"
    if add_qseq:
        table_k += " qseq"

    # SHORT SEQUENCES: -task 'blastp-short' -word_size 2 -seg 'no' -evalue 20000
    if len(sequence) <= 15:
        os.system(
            blast_path
            + "blastp -query "
            + tmp_folder
            + "sequence_temp.txt -db "
            + database
            + " -matrix PAM30 -task 'blastp-short' -word_size 2 -seg 'no' -evalue 20000 -out "
            + blast_filename
            + f" -outfmt '{table_k}'"
            + ungapped
            + " -max_target_seqs "
            + str(max_target_seqs)
            + " -num_threads "
            + str(n_cpu)
        )
    else:
        os.system(
            blast_path
            + "blastp -query "
            + tmp_folder
            + "sequence_temp.txt -db "
            + database
            + " -out "
            + blast_filename
            + f" -outfmt '{table_k}'"
            + ungapped
            + " -max_target_seqs "
            + str(max_target_seqs)
            + " -num_threads "
            + str(n_cpu)
        )
    resultf = open(blast_filename, mode="r").read().splitlines()

    results = []
    if len(resultf) == 0:
        return None
    for line in resultf:
        tmp = line.split("\t")
        numb_list = [tmp[0]] + list(map(lambda x: float(x), tmp[1:]))
        results.append(numb_list)
    if delete_blast_file:
        os.system("rm -f " + blast_filename)
    os.system("rm -f " + tmp_folder + "sequence_temp.txt")
    del resultf
    return results


def process_blast_results(
    blast_results_list,
    len_of_query,
    return_dictionary_of_numpy_matrices=False,
    filter_best_match_for_same_subject=True,
    add_identity_on_shortest=True,
):
    """
    If filter_best_match_for_same_subject is True than only the first among the possible alignments of the query
    with that subject sequence is kept. Blast results are sorted according to subject sequences and then to
    evalue
    :param blast_results_list:
    :param len_of_query:
    :param return_dictionary_of_numpy_matrices:
    :param filter_best_match_for_same_subject:
    :param add_identity_on_shortest:
    :return:
    """
    if not isinstance(len_of_query, int):
        print(
            f"WARNING in {process_blast_results.__name__} type(len_of_query) is not Int, "
            "probably screwing results up!!"
        )
    if blast_results_list is None:
        return ([], {}) if return_dictionary_of_numpy_matrices else []
    if (
        not filter_best_match_for_same_subject
        and not add_identity_on_shortest
        and not return_dictionary_of_numpy_matrices
    ):
        return blast_results_list
    last_subj = ""
    new_results = []
    if return_dictionary_of_numpy_matrices:
        subjects_dict = {}
    for res in blast_results_list:
        if return_dictionary_of_numpy_matrices:
            if res[0] not in subjects_dict:
                subjects_dict[res[0]] = np.array(
                    res[1:], dtype=np.float64, ndmin=2
                )  # this removes the ID
            else:
                subjects_dict[res[0]] = np.vstack(
                    (
                        subjects_dict[res[0]],
                        np.array(res[1:], dtype=np.float64),
                    )
                )
        if filter_best_match_for_same_subject and list(res)[0] == last_subj:
            continue
        if add_identity_on_shortest:
            den = 1.0 * min([res[-1], len_of_query])
            new_results.append(res + [100.0 * res[-2] / den])
        else:
            new_results.append(res)
        last_subj = res[0]
    if return_dictionary_of_numpy_matrices:
        return new_results, subjects_dict
    return new_results


def generate_random_sets(
    clusters,
    number_of_sets=10,
    fraction_in_test=0.1,
    train_set_key="LR",
    test_set_key="TS",
    early_break=True,
):
    """
    Generate random sets from clusters.

    :param clusters: input clusters
    :param number_of_sets: total number of sets to generate
    :param fraction_in_test: fraction of sequences to be put in test
    :param train_set_key: Prefix to add to train set
    :param test_set_key: Prefix to add to test set
    :param early_break: if True, break after a suitable combination of training-test set is generated to feed the model
    :return:
    """
    if type(clusters) is str and os.path.isfile(clusters):
        clusters = open(clusters, mode="r").read().splitlines()
        clusters = [cl.split() for cl in clusters]
    num_ids = 0
    c_sizes = []
    for c in clusters:
        c_sizes += [len(c)]
        num_ids += len(c)
    print(f"{num_ids} ids in {len(clusters)} clusters")
    random.seed(a=42)  # for reproducibility
    n = 0
    in_other_tests = []
    while n < number_of_sets:
        train_set_name = f"{train_set_key}_{n:02}"
        test_set_name = f"{test_set_key}_{n:02}"
        with open(train_set_name, mode="w") as train:
            with open(test_set_name, mode="w") as test:
                ids_in_test = []
                labs_in_test = []
                test_rel_size = 0.0
                seq_in_test = 0
                while test_rel_size < fraction_in_test:
                    cl_id = int(random.random() * len(clusters))
                    while cl_id in in_other_tests:
                        cl_id = int(random.random() * len(clusters))
                    in_other_tests += [cl_id]
                    if len(in_other_tests) == len(clusters):
                        print(
                            f"\nWARNING in {generate_random_sets.__name__} we have already generated test files"
                            " with all clusters.\nNew test files will share clusters with existing ones...\n"
                            f"Starting from {n} included...\n"
                        )
                        in_other_tests = []
                    ids_in_test += [cl_id]
                    for idt in clusters[cl_id]:
                        print(f"{str(idt)}", file=test)
                        labs_in_test.append(idt)
                    test_rel_size += 1.0 * c_sizes[cl_id] / float(num_ids)
                    seq_in_test += c_sizes[cl_id]
                for j in range(len(clusters)):
                    if j not in ids_in_test:
                        for idt in clusters[j]:
                            print(f"{str(idt)}", file=train)
                print(
                    f"Generated set {train_set_name} and {test_set_name} - {seq_in_test} sequences"
                    f" for {len(ids_in_test)} clusters in test, among ({num_ids}, {len(clusters)}) in total."
                    f" Test relative size is {test_rel_size}"
                )
        if abs(test_rel_size - fraction_in_test) < 0.05 and early_break:
            return labs_in_test
        n += 1


def train_and_test_from_fasta(
    fasta_file: str,
    number_of_sets: int = 10,
    fraction_in_test: float = 0.1,
    cluster_file_name: str = "s2d_clusters.txt",
    seq_id_low_thresh: float = 25.0,
    use_very_loose_condition: bool = False,
    n_cpu: int = 2,
    max_target_seqs: int = 200,
    delete_db_when_done: bool = True,
    train_set_key: str = "LR",
    test_set_key: str = "TS",
    early_break: bool = True,
):
    """
    Use a fasta database to generate test and training sets based on user defined parameters.

    :param fasta_file: Fasta file name.
    :type fasta_file: str
    :param number_of_sets: Number of sets to generate.
    :type number_of_sets: int
    :param fraction_in_test: Fraction of proteins to be put in test set.
    :type fraction_in_test: float
    :param cluster_file_name: Name of the cluster file.
    :type cluster_file_name: str
    :param seq_id_low_thresh: Sequence identity lower threshold.
    :type seq_id_low_thresh: float
    :param use_very_loose_condition: ignore
    :type use_very_loose_condition: bool
    :param n_cpu: Number of cpus.
    :type n_cpu: int
    :param max_target_seqs: Maximum number of target sequences.
    :type max_target_seqs: int
    :param delete_db_when_done: if ``True`` delete the blast database when done.
    :type delete_db_when_done: bool
    :param train_set_key: Prefix to append to training files.
    :type train_set_key: str
    :param test_set_key: Prefix to append to test files.
    :type test_set_key: str
    :param early_break: if ``True``, break after a suitable combination of training-test set is generated to feed the model.
    :type early_break: bool
    :return: save clusters in files (or return the 1st suitable one).
    """
    if not 0 < fraction_in_test < 1:
        print(
            f"ERROR in {train_and_test_from_fasta.__name__}, fraction_in_test has to be in [0,1] now {fraction_in_test}"
        )

    print(
        "Clustering proteins by sequence identity\n"
        "----------------------------------------\n"
        "Parameters:\n"
        f"    - Number of sets to split: {number_of_sets}\n"
        f"    - Sequence identity low threshold: {seq_id_low_thresh}%\n"
        f"    - Fraction of sequences in test set: {fraction_in_test}"
    )
    print("\n\n*** Creating sequences pairs for clustering")
    pairs = create_pairs_for_clustering(
        fasta_file,
        seq_id_low_thresh=seq_id_low_thresh,
        use_very_loose_condition=use_very_loose_condition,
        out_file=None,
        n_cpu=n_cpu,
        max_target_seqs=max_target_seqs,
        delete_blast_databases_when_done=delete_db_when_done,
    )
    print("*** Clustering from pairs of sequences")
    clusters = clustering_from_pairs(pairs, out_file=cluster_file_name)
    print("*** Generating sets")
    if early_break:
        return generate_random_sets(
            clusters,
            number_of_sets=number_of_sets,
            fraction_in_test=fraction_in_test,
            train_set_key=train_set_key,
            test_set_key=test_set_key,
            early_break=early_break,
        )

    generate_random_sets(
        clusters,
        number_of_sets=number_of_sets,
        fraction_in_test=fraction_in_test,
        train_set_key=train_set_key,
        test_set_key=test_set_key,
        early_break=early_break,
    )
    return



"""Utility Functions for working with Small Molecule Graphs in RDKit.

Many of these utilities are adapted from useful_rdkit_utils (https://github.com/PatWalters/useful_rdkit_utils) by Pat Walters.

Junction tree code adapted from Wengong Jin https://github.com/wengong-jin/icml18-jtnn
"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT

# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from __future__ import annotations

import collections
from typing import Any, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree

from graphein.utils.dependencies import import_message, requires_python_libs

try:
    import rdkit
    from rdkit import Chem as Chem
    from rdkit import DataStructs
    from rdkit.Chem import AllChem
    from rdkit.Chem.Descriptors import (
        TPSA,
        MolLogP,
        MolWt,
        NumHAcceptors,
        NumHDonors,
    )
    from rdkit.Chem.Descriptors3D import NPR1, NPR2
    from rdkit.Chem.rdMolTransforms import ComputeCentroid
except ImportError:
    import_message(
        "graphein.molecule.utils", "rdkit", "rdkit", True, extras=True
    )

try:
    import selfies as sf
except ImportError:
    import_message(
        "graphein.molecule.utils", "selfies", None, True, extras=True
    )


MST_MAX_WEIGHT: int = 100
MAX_NCAND: int = 2000


@requires_python_libs("rdkit")
def get_center(
    mol: Union[nx.Graph, Chem.Mol], weights: Optional[np.ndarray] = None
) -> np.ndarray:
    """Compute the centroid of the conformation.

    Hydrogens are ignored and no attention is paid to the difference in sizes
    of the heavy atoms; however, an optional vector of weights can be passed.

    :param mol: Molecular Graph or RDkit Mol to compute the center of.
    :type mol: Union[nx.Graph, Chem.Mol]
    :return: The centroid of the molecule.
    :rtype: np.ndarray
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    assert (
        mol.GetNumConformers() > 0
    ), "Molecule must have at least one conformer."
    return np.array(ComputeCentroid(mol.GetConformer(0), weights=weights))


@requires_python_libs("rdkit")
def get_shape_moments(mol: Union[nx.Graph, Chem.Mol]) -> Tuple[float, float]:
    """Calculate principal moments of inertia as defined in https://pubs.acs.org/doi/10.1021/ci025599w

    :param mol: Molecular Graph or RDkit Mol to compute the moments of intertia of.
    :type mol: Union[nx.Graph, Chem.Mol]
    :return: First 2 moments as a tuple.
    :rtype: Tuple[float, float]
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    assert (
        mol.GetNumConformers() > 0
    ), "Molecule must have at least one conformer."
    npr1 = NPR1(mol)
    npr2 = NPR2(mol)
    return npr1, npr2


@requires_python_libs("rdkit")
def count_fragments(mol: Union[nx.Graph, Chem.Mol]) -> int:
    """Counts the number of the disconnected fragments in a molecule.

    :param mol: The molecule or molecular graph.
    :type mol: Union[nx.Graph, Chem.Mol]
    :return: Number of fragments.
    :rtype: int
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    return len(Chem.GetMolFrags(mol, asMols=True))


@requires_python_libs("rdkit")
def get_max_ring_size(mol: Union[nx.Graph, Chem.Mol]) -> int:
    """
    Get the size of the largest ring in a molecule.

    :param mol: Input Mol or molecular graph.
    :type mol: Union[nx.Graph, Chem.Mol]
    :return: size of the largest ring or 0 for an acyclic molecule.
    :rtype: int
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]

    ri = mol.GetRingInfo()
    atom_rings = ri.AtomRings()
    return 0 if len(atom_rings) == 0 else max(len(x) for x in ri.AtomRings())


@requires_python_libs("rdkit")
def label_rdmol_atoms(
    mol: Union[nx.Graph, Chem.Mol], labels: List[Any]
) -> Union[nx.Graph, Chem.Mol]:
    """Sets an atomNote label on each atom in the underlying RDKit Mol.

    :param mol: RDkit Mol or molecular graph containing a mol.
    :type mol: Union[nx.Graph, Chem.Mol]
    :param labels: A list of labels to set on each atom.
    :type labels: _type_
    :return: _description_
    :rtype: Union[nx.Graph, Chem.Mol]
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]

    [atm.SetProp("atomNote", "") for atm in mol.GetAtoms()]
    for atm in mol.GetAtoms():
        idx = atm.GetIdx()
        mol.GetAtomWithIdx(idx).SetProp("atomNote", f"{labels[idx]}")
    return mol


@requires_python_libs("rdkit")
def tag_rdmol_atoms(
    mol, atoms_to_tag, tag: str = "x"
) -> Union[nx.Graph, Chem.Mol]:
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    [atm.SetProp("atomNote", "") for atm in mol.GetAtoms()]
    [mol.GetAtomWithIdx(idx).SetProp("atomNote", tag) for idx in atoms_to_tag]
    return mol


@requires_python_libs("rdkit")
def get_mol(smiles: str) -> rdkit.Chem.rdchem.Mol:
    """
    Function for getting rdmol from smiles. Applies kekulization.

    :param smiles: smiles string to get
    :type x: str
    :return: rdmol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    Chem.Kekulize(mol)
    return mol


@requires_python_libs("rdkit")
def get_smiles(mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol]) -> str:
    """
    Function for getting smiles from rdmol. Applies kekulization.

    :param mol: rdmol to get
    :type mol: Union[rdkit.Chem.rdchem.Mol, nx.Graph]
    :return: smiles string
    :rtype: str
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    return Chem.MolToSmiles(mol, kekuleSmiles=True)


@requires_python_libs("rdkit")
def sanitize(mol: rdkit.Chem.rdchem.Mol) -> rdkit.Chem.rdchem.Mol:
    """
    Function for sanitizing a rdmol

    :param mol: rdmol to sanitize
    :type mol: rdkit.Chem.rdchem.Mol
    :return: sanitized rdmol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    try:
        smiles = get_smiles(mol)
        mol = get_mol(smiles)
    except Exception as e:
        return None
    return mol


@requires_python_libs("rdkit")
def copy_edit_mol(mol: rdkit.Chem.rdchem.Mol) -> rdkit.Chem.rdchem.Mol:
    """
    Function for copying a rdmol

    :param mol: rdmol to copy
    :type mol: rdkit.Chem.rdchem.Mol
    :return: copied rdmol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    new_mol = Chem.RWMol(Chem.MolFromSmiles(""))
    for atom in mol.GetAtoms():
        new_atom = copy_atom(atom)
        new_mol.AddAtom(new_atom)
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom().GetIdx()
        a2 = bond.GetEndAtom().GetIdx()
        bt = bond.GetBondType()
        new_mol.AddBond(a1, a2, bt)
    return new_mol


@requires_python_libs("rdkit")
def get_clique_mol(mol: rdkit.Chem.rdchem.Atom, atoms: List[int]):
    """
    Function for getting clique rdmol

    :param mol: rdmol to get clique
    :type mol: rdkit.Chem.rdchem.Mol
    :param atoms: atoms to cut
    :type atoms: List[int]
    :return: clique rdmol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    smiles = Chem.MolFragmentToSmiles(mol, atoms, kekuleSmiles=True)
    new_mol = Chem.MolFromSmiles(smiles, sanitize=False)
    new_mol = copy_edit_mol(new_mol).GetMol()
    new_mol = sanitize(new_mol)  # We assume this is not None
    return new_mol


@requires_python_libs("rdkit")
def copy_rdmol_atom(atom: rdkit.Chem.rdchem.Atom) -> rdkit.Chem.rdchem.Atom:
    """
    Function for copying an atom

    :param mol: atom to copy
    :type mol: rdkit.Chem.rdchem.Atom
    :return: copied atom
    :rtype: rdkit.Chem.rdchem.Atom
    """
    new_atom = Chem.Atom(atom.GetSymbol())
    new_atom.SetFormalCharge(atom.GetFormalCharge())
    new_atom.SetAtomMapNum(atom.GetAtomMapNum())
    return new_atom


@requires_python_libs("rdkit")
def get_morgan_fp(
    mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol],
    radius: int = 2,
    n_bits: int = 20,
) -> rdkit.DataStructs.cDataStructs.ExplicitBitVect:
    """
    Function for getting morgan fingerprint from an RDkit molecule or molecular graph.

    :param mol: RDKit molecule or molecular graph.
    :type mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol]
    :param radius: Fingerprint radius.
    :type radius: int
    :param n_bits: Number of bits.
    :type n_bits: int
    :return: Morgan fingerprint.
    :rtype: rdkit.Chem.rdMolDescriptors.MorganFingerprint
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    return AllChem.GetMorganFingerprintAsBitVect(
        mol, radius=radius, nBits=n_bits
    )


@requires_python_libs("rdkit")
def get_morgan_fp_np(
    mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol],
    radius: int = 2,
    n_bits: int = 20,
) -> rdkit.Chem.rdMolDescriptors.MorganFingerprint:
    """
    Function for getting morgan fingerprint from an RDkit molecule or molecular graph.

    :param mol: RDKit molecule or molecular graph.
    :type mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol]
    :param radius: Fingerprint radius.
    :type radius: int
    :param n_bits: Number of bits.
    :type n_bits: int
    :return: Morgan fingerprint.
    :rtype: rdkit.Chem.rdMolDescriptors.MorganFingerprint
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]

    arr = np.zeros((0,), dtype=np.int8)
    fp = get_morgan_fp(mol=mol, radius=radius, n_bits=n_bits)
    DataStructs.ConvertToNumpyArray(fp, arr)
    return arr


@requires_python_libs("rdkit")
def compute_fragments(mol: Union[nx.Graph, Chem.Mol]) -> List[Chem.Mol]:
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    return list(Chem.GetMolFrags(mol, asMols=True))


@requires_python_libs("rdkit")
def get_mol_weight(mol: Union[nx.Graph, Chem.Mol]) -> float:
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]
    return mol  # TDOO


@requires_python_libs("rdkit")
def get_qed_score(
    mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol],
) -> Union[float, None]:
    """Computes the Quantitative Estimate of Druglikeness (QED) score for a molecule or molecular graph.

        Quantifying the chemical beauty of drugs
        G. Richard Bickerton, Gaia V. Paolini, Jérémy Besnard, Sorel Muresan & Andrew L. Hopkins
        Nature Chemistry volume 4, pages90–98 (2012

    :param mol: Molecule or molecular graph.
    :type mol: Union[nx.Graph, rdkit.Chem.rdchem.Mol]
    :return: QED score.
    :rtype: float or None if an exception is encountered
    """
    if isinstance(mol, nx.Graph):
        mol = mol.graph["rdmol"]

    try:
        return Chem.QED.qed(mol)
    except Exception as e:
        # log.warning(e)
        return None


def simplify_smile(smile: str) -> str:
    """
    Simplifies a SMILE string by removing hydrogen atoms (``H``),
    chiral specifications (``'@'``), charges (``+`` / ``-``), ``'#'``-characters,
    and square brackets (``'['``, ``']'``).

    :param smile_str: A smile string, e.g., ``C[C@H](CCC(=O)NCCS(=O)(=O)[O-])``
    :type smile_str: str
    :returns: Returns a simplified SMILE string, e.g., ``CC(CCC(=O)NCCS(=O)(=O)O)``.
    :rtype: str
    """
    remove_chars = ["@", "-", "+", "H", "[", "]", "#"]
    stripped_smile = []
    for sym in smile:
        if sym.isalpha():
            sym = sym.upper()
        if sym not in remove_chars:
            stripped_smile.append(sym)
    return "".join(stripped_smile)


@requires_python_libs("selfies")
def smile_to_selfies(smile: str) -> str:
    """Encodes a SMILES string into a Selfies string.

    :param smile: A valid SMILES string. E.g. ``"C1=CC=CC=C1"``.
    :type smile: str
    :return: Selfies string. E.g. ``"[C][=C][C][=C][C][=C][Ring1][=Branch1]"``.
    :rtype: str
    """
    return sf.encoder(smile)


@requires_python_libs("selfies")
def selfies_to_smile(selfie: str) -> str:
    """Decodes a selfies string into a SMILES string.

    :param selfie: The selfies string to decode. E.g. ``"[C][=C][C][=C][C][=C][Ring1][=Branch1]"``.
    :type selfie: str
    :return: The decoded SMILES string. E.g. ``"C1=CC=CC=C1"``.
    :rtype: str
    """
    return sf.decoder(selfie)


@requires_python_libs("rdkit")
def tree_decomp(mol: rdkit.Chem.rdchem.Mol) -> Tuple[List]:
    """
    Function for decomposing rdmol to a tree

    :param mol: rdmol to decompose
    :type mol: rdkit.Chem.rdchem.Mol
    :return: decomposed cliques and edges between them
    :rtype: Tuple[List]
    """
    n_atoms = mol.GetNumAtoms()
    if n_atoms == 1:  # special case
        return [[0]], []

    cliques = []
    for bond in mol.GetBonds():
        a1 = bond.GetBeginAtom().GetIdx()
        a2 = bond.GetEndAtom().GetIdx()
        if not bond.IsInRing():
            cliques.append([a1, a2])

    ssr = [list(x) for x in Chem.GetSymmSSSR(mol)]
    cliques.extend(ssr)

    nei_list = [[] for _ in range(n_atoms)]
    for i in range(len(cliques)):
        for atom in cliques[i]:
            nei_list[atom].append(i)

    # Merge Rings with intersection > 2 atoms
    for i in range(len(cliques)):
        if len(cliques[i]) <= 2:
            continue
        for atom in cliques[i]:
            for j in nei_list[atom]:
                if i >= j or len(cliques[j]) <= 2:
                    continue
                inter = set(cliques[i]) & set(cliques[j])
                if len(inter) > 2:
                    cliques[i].extend(cliques[j])
                    cliques[i] = list(set(cliques[i]))
                    cliques[j] = []

    cliques = [c for c in cliques if len(c) > 0]
    nei_list = [[] for _ in range(n_atoms)]
    for i in range(len(cliques)):
        for atom in cliques[i]:
            nei_list[atom].append(i)

    # Build edges and add singleton cliques
    edges = collections.defaultdict(int)
    for atom in range(n_atoms):
        if len(nei_list[atom]) <= 1:
            continue
        cnei = nei_list[atom]
        bonds = [c for c in cnei if len(cliques[c]) == 2]
        rings = [c for c in cnei if len(cliques[c]) > 4]
        if len(bonds) > 2 or (
            len(bonds) == 2 and len(cnei) > 2
        ):  # In general, if len(cnei) >= 3, a singleton should be added, but 1 bond + 2 ring is currently not dealt with.
            cliques.append([atom])
            c2 = len(cliques) - 1
            for c1 in cnei:
                edges[(c1, c2)] = 1
        elif len(rings) > 2:  # Multiple (n>2) complex rings
            cliques.append([atom])
            c2 = len(cliques) - 1
            for c1 in cnei:
                edges[(c1, c2)] = MST_MAX_WEIGHT - 1
        else:
            for i in range(len(cnei)):
                for j in range(i + 1, len(cnei)):
                    c1, c2 = cnei[i], cnei[j]
                    inter = set(cliques[c1]) & set(cliques[c2])
                    edges[(c1, c2)] = max(edges[(c1, c2)], len(inter))
    edges = [u + (MST_MAX_WEIGHT - v,) for u, v in edges.items()]
    if not edges:
        return cliques, edges

    # Compute Maximum Spanning Tree
    row, col, data = zip(*edges)
    n_clique = len(cliques)
    clique_graph = csr_matrix((data, (row, col)), shape=(n_clique, n_clique))
    junc_tree = minimum_spanning_tree(clique_graph)
    row, col = junc_tree.nonzero()
    edges = [(row[i], col[i]) for i in range(len(row))]
    return (cliques, edges)



"""Tests for graphein.protein.tensor.geometry."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import numpy as np
import pytest

import graphein.protein.tensor as gpt
from graphein.protein.tensor.angles import (
    get_backbone_bond_angles,
    get_backbone_bond_lengths,
)
from graphein.protein.tensor.geometry import (
    IDEAL_BB_BOND_ANGLES,
    IDEAL_BB_BOND_LENGTHS,
    center_protein,
    kabsch,
    quaternion_to_matrix,
)

try:
    import torch

    TORCH_AVAIL = True
except ImportError:
    TORCH_AVAIL = False


def test_center_protein():
    x = torch.randn((5, 3))
    x_centered = center_protein(x)
    assert torch.allclose(x_centered.mean(dim=0), torch.zeros(3), atol=1e-5)


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_whole_protein_kabsch():
    # Compute the RMSD between the original and aligned coordinates
    def rmsd(x, y):
        return torch.sqrt(torch.mean((x - y) ** 2))

    # Test 2D
    A = torch.rand(10, 2)
    R0 = torch.tensor(
        [[np.cos(60), -np.sin(60)], [np.sin(60), np.cos(60)]],
        dtype=torch.float,
    )
    B = (R0.mm(A.T)).T
    t0 = torch.tensor([3.0, 3.0])
    B += t0
    R, t = kabsch(A, B, return_transformed=False)
    A_aligned = (R.mm(A.T)).T + t
    error = rmsd(A_aligned, B)
    assert error < 1e-6, "RMSD too high after alignment"

    # Test 3D
    coords = torch.rand(10, 3)
    # Create a random rotation matrix
    quat = torch.rand(4)
    rot = quaternion_to_matrix(quat)
    # Apply a random rotation and translation to coords
    coords_distorted = rot.mm(coords.T).T + torch.rand(3) * 4
    r, t = kabsch(coords_distorted, coords, return_transformed=False)
    coords_aligned = r.mm(coords_distorted.T).T + t

    assert rmsd(coords_aligned, coords) < 1e-6, "RMSD too high after alignment"
    assert torch.allclose(
        coords_aligned, coords, atol=1e-6
    ), "Coords differ too much after alignment"


def test_idealise_backbone():
    protein = gpt.Protein().from_pdb_code(pdb_code="5caj", chain_selection="A")
    protein.coords = protein.coords[:20, :, :]

    ideal = protein.idealize_backbone(n_iter=100)

    IDEAL_BL = torch.tensor(
        IDEAL_BB_BOND_LENGTHS, device=protein.coords.device
    )
    IDEAL_BA = torch.tensor(IDEAL_BB_BOND_ANGLES, device=protein.coords.device)

    ideal_bl = IDEAL_BL - get_backbone_bond_lengths(ideal[:, :4, :])
    true_bl = IDEAL_BL - get_backbone_bond_lengths(protein.coords[:, :4, :])

    ideal_ba = IDEAL_BA - get_backbone_bond_angles(ideal[:, :4, :])
    true_ba = IDEAL_BA - get_backbone_bond_angles(protein.coords[:, :4, :])

    assert torch.mean(torch.abs(ideal_ba)) < torch.mean(
        torch.abs(true_ba)
    ), "Bond angles not more idealised"
    assert torch.mean(torch.abs(ideal_bl)) < torch.mean(
        torch.abs(true_bl)
    ), "Bond lengths not more idealised"



"""Types for graphein.protein.tensor.

Graphein provides many types for commonly used tensors of specific shapes.
"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import NewType, Optional, Union

from jaxtyping import Float, Int
from torch import Tensor

# Positions
AtomTensor = NewType("AtomTensor", Float[Tensor, "residues 37 3"])
"""
``torch.float[-1, 37, 3]``

Tensor of atom coordinates. The first dimension is the length of the protein,
the second the number of canonical atom types. The last dimension contains the
x,y,z positions.

.. seealso:: :class:`ResidueTensor` :class:`CoordTensor`
"""

BackboneTensor = NewType("BackboneTensor", Float[Tensor, "residues 4 3"])
"""
``torch.float[-1, 4, 3]``

Tensor of backbone atomic coordinates. The first dimension is the length of the
protein (or batch), the second dimension corresponds to ``[N, Ca, C, O]`` atoms
and the last dimension contains the x,y,z coordinates.

.. seealso:: :ref:`graphein.protein.tensor.types.AtomTensor

"""

ResidueTensor = NewType("ResidueTensor", Float[Tensor, "37 3"])
"""
``torch.float[37, 3]``

Tensor of atom coordinates for a residue. Each index in dimension one
corresponds to an atom (See: :ref:`graphein.protein.resi_atoms.ATOM_NUMBERING`).
The last dimension contains the x,y,z positions. Missing atoms are denoted by a
fill value (default = ``1e-5``).

.. seealso:: :class:`AtomTensor` :class:`CoordTensor`
"""


CoordTensor = NewType("CoordTensor", Float[Tensor, "nodes 3"])
"""
``torch.float[-1, 3]``

Tensor of coordinates. The first dimension is the length of the protein
(or batch of proteins), the last dimension contains the x,y,z positions."""

PositionTensor = NewType("PositionTensor", Union[AtomTensor, CoordTensor])
"""
Union[:ref:`graphein.protein.tensor.types.AtomTensor,
:ref:`graphein.protein.tensor.types.CoordTensor]


Union of ``AtomTensor`` and ``CoordTensor``.

.. seealso:: :class:`AtomTensor`, :class:`CoordTensor`
"""

# Represenations
BackboneFrameTensor = NewType(
    "BackboneFrameTensor", Float[Tensor, "residues 3 3"]
)
"""
``torch.float[-1, 3, 3]``

Tensor of backbone frames as rotation matrices. The first dimension is the
length of the protein, the second and third dimensions specify a rotation matrix
relative to an idealised residue.

.. seealso::
    :ref:`graphein.protein.tensor.reconstruction.get_ideal_backbone_coords
"""

ResidueFrameTensor = NewType("ResidueFrameTensor", Float[Tensor, "3 3"])
"""
``torch.float[3, 3]``

.. seealso:: :class:`BackboneFrameTensor`
"""


# Rotations
EulerAngleTensor = NewType("EulerAngleTensor", Float[Tensor, "nodes 3"])

QuaternionTensor = NewType("QuaternionTensor", Float[Tensor, "nodes 4"])
"""
``torch.float[-1, 4]``

Tensor of quaternions. The first dimension is the length of the protein
(or batch), the second dimension is the quaternion.

.. see:: https://en.wikipedia.org/wiki/Quaternion
"""


TransformTensor = NewType("TransformTensor", Float[Tensor, "nodes 4 4"])


RotationMatrix2D = NewType("RotationMatrix2D", Float[Tensor, "2 2"])
"""
``torch.float[2, 2]``

Specifies a 2D rotation matrix.

.. seealso:: :class:`RotationMatrix3D` :class:`RotationMatrix` :class:`RotationTensor`
"""

RotationMatrix3D = NewType("RotationMatrix3D", Float[Tensor, "3 3"])
"""
``torch.float[3, 3]``

Specifies a 3D rotation matrix.

.. seealso:: :class:`RotationMatrix2D` :class:`RotationMatrix` :class:`RotationTensor`
"""

RotationMatrix = NewType(
    "RotationMatrix", Union[RotationMatrix2D, RotationMatrix3D]
)
"""
``Union[RotationMatrix3D, RotationMatrix2D]``

Specifies a rotation matrix in either 2D or 3D.

.. seealso:: :class:`RotationMatrix2D` :class:`RotationMatrix3D` :class:`RotationTensor`
"""


RotationMatrixTensor = NewType(
    "RotationMatrixTensor", Float[Tensor, "nodes 3 3"]
)

RotationTensor = NewType(
    "RotationTensor", Union[QuaternionTensor, RotationMatrixTensor]
)


# Angles
DihedralTensor = NewType(
    "DihedralTensor",
    Union[Float[Tensor, "residues 3"], Float[Tensor, "residues 6"]],
)
"""
``Union[torch.float[-1, 3], torch.float[-1, 6]]``

Tensor of backbone dihedral angles (phi, psi, omega). Either in degrees/radians
or embedded on the unit sphere ``[cos(phi), sin(phi), cos(psi), sin(psi), ...]``

.. seealso::

    :meth:graphein.protein.tensor.angles.dihedrals,
    :meth:graphein.protein.tensor.angles.dihedrals_to_rad
    :meth:graphein.protein.tensor.plot.plot_dihedrals

"""

TorsionTensor = NewType(
    "TorsionTensor",
    Union[Float[Tensor, "residues 4"], Float[Tensor, "residues 8"]],
)
"""
``Union[torch.float[-1, 4], torch.float[-1, 8]]``

Tensor of sidechain torsion angles ``[chi1, chi2, chi3, chi4]``. Either in
degrees/radians or embedded on the unit sphere:
``[cos(chi1), sin(chi1), cos(chi2), sin(chi2), ...]``.

.. seealso::

    :meth:graphein.protein.tensor.angles.sidechain_torsions,
    :meth:graphein.protein.tensor.angles.torsions_to_rad

"""

BackboneFrameTensor = NewType(
    "BackboneFrameTensor", Float[Tensor, "residues 3 3"]
)
"""
``torch.float[-1, 3, 3]``

Tensor of backbone frames as rotation matrices. The first dimension is the
length of the protein, the second and third dimensions specify a rotation matrix
relative to an idealised residue.

.. seealso::

    :meth:`graphein.protein.tensor.reconstruction.get_ideal_backbone_coords`
    :meth:`graphein.protein.tensor.representation.get_backbone_frames`
"""

ResidueFrameTensor = NewType("ResidueFrameTensor", Float[Tensor, "3 3"])
"""
``torch.float[-1, 3, 3]``

.. seealso:: :class:`BackboneFrameTensor`
"""

EdgeTensor = NewType("EdgeTensor", Int[Tensor, "2 edges"])


OrientationTensor = NewType("OrientationTensor", Float[Tensor, "nodes 2 3"])


ScalarTensor = NewType("ScalarTensor", Float[Tensor, "nodes"])

OptTensor = NewType("OptTensor", Optional[Tensor])



# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from pydantic import BaseModel


class TRRUSTConfig(BaseModel):
    """
    Config object for parsing gene regulatory networks from TRRUST: https://www.grnpedia.org/trrust/

    :param filtering_functions: List of functions to apply to the the TRRUST dataframe prior to graph construction. Defaults to None
    :type filtering_functions: List[Callable], optional
    :param root_dir: Specifies location of TRRUST dataset (will download to this path if not available). Defaults to None.
    :type root_dir: pathlib.Path, optional
    :param kwargs:
    :type kwargs: Dict[str, Union[str, int, float]], optional
    """

    filtering_functions: Optional[List[Callable]] = None
    root_dir: Optional[Path] = None
    kwargs: Optional[Dict[str, Union[str, int, float]]] = None


class RegNetworkConfig(BaseModel):
    """Config object containing parameters for parsing gene regulatory networks from RegNetwork: http://regnetworkweb.org/.

    :param filtering_functions: List of functions to apply to the the RegNetwork dataframe prior to graph construction. Defaults to None
    :type filtering_functions: List[Callable], optional


    """

    filtering_functions: Optional[List[Callable]] = None
    root_dir: Optional[Path] = None
    kwargs: Optional[Dict[str, Union[str, int, float]]] = None


class GRNGraphConfig(BaseModel):
    """Config object for gene regulatory network graph construction.

    :param kwargs: Keyword args for GRN graph construction
    :type kwargs: Dict[str, Union[str, int, float]], optional
    :param trrust_config: Config object specifying parameters for parsing TRRUST. Defaults to default config object.
    :type trrust_config: graphein.grn.config.TRRUSTConfig, optional
    :param regnetwork_config: Config object specifying parameters for parsing RegNetwork. Defaults to default config object.
    :type regnetwork_config: graphein.grn.config.RegNetworkConfig, optional
    """

    kwargs: Optional[Dict[str, Union[str, int, float]]] = {}
    trrust_config: Optional[TRRUSTConfig] = TRRUSTConfig()
    regnetwork_config: Optional[RegNetworkConfig] = RegNetworkConfig()

    class Config:
        arbitrary_types_allowed: bool = True



"""Plotting utilities for protein tensors.

All plots are produced with Plotly and so are loggable to Weights and Biases.
"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Any, Dict, List, Optional, Union

import plotly.express as px
import plotly.graph_objects as go
import torch

from ..resi_atoms import ATOM_NUMBERING
from .angles import dihedrals_to_rad
from .representation import get_full_atom_coords
from .types import AtomTensor, CoordTensor, DihedralTensor


def plot_dihedrals(
    dihedrals: DihedralTensor, to_rad: bool = True
) -> go.Figure:
    """Plots a heatmap of dihedral angles.

    .. code-block:: python
        import torch

        x = torch.rand((32, 6))
        plot_dihedrals(x)


    .. seealso::
        :meth:`graphein.protein.tensor.angles.dihedrals``
        :meth:`graphein.protein.tensor.angles.dihedrals_to_rad`
        :class:`graphein.protein.tensor.types.DihedralTensor`

    :param dihedrals: Tensor of Dihedral Angles
    :type dihedrals: graphein.protein.tensor.DihedralTensor
    :param to_rad: Whether or not to convert to radians, defaults to ``True``
    :type to_rad: bool, optional
    :returns: Plotly figure of dihedral angles (``px.imshow``)
    :rtype: go.Figure
    """
    if dihedrals.shape[1] == 6 and to_rad:
        dihedrals = dihedrals_to_rad(dihedrals, concat=True)

    return px.imshow(dihedrals, aspect="auto")  # , vmin=-np.pi, vmax=np.pi)


def plot_distance_matrix(
    x: Union[CoordTensor, AtomTensor], **kwargs: Dict[str, Any]
) -> go.Figure:
    """
    Computes and plots a distance matrix of a ``CoordTensor`` or ``AtomTensor``.

    :param x: Tensor of structure to plot
    :type x: Union[CoordTensor, AtomTensor]
    :return: Plotly figure of distance matrix (``px.imshow``)
    :rtype: go.Figure
    """
    if x.ndim == 3:
        x = get_full_atom_coords(x, **kwargs)

    dist_mat = torch.cdist(x, x)
    return px.imshow(dist_mat)


def plot_structure(
    x: Union[CoordTensor, AtomTensor],
    atoms: List[str] = ["N", "CA", "C", "O", "CB"],
    lines: bool = True,
    residue_ids: Optional[List[str]] = None,
) -> go.Figure:
    """Plots a protein structure in 3D.

    :param x: Coordinates of the protein. Either an AtomTensor
        ``(Length x Num Atom Types x 3)`` or a CoordTensor ``(Length x 3)``.
    :type x: Union[CoordTensor, AtomTensor]
    :param atoms: List of atoms to include in the plot,
        defaults to ``["N", "CA", "C", "O", "CB"]``.
    :type atoms: List[str], optional
    :param lines: Whether or not to join points with lines,
        defaults to ``True``.
    :type lines: bool, optional
    :return: Figure object
    :rtype: go.Figure
    """

    mode = "lines" if lines else "markers"
    if x.ndim == 2:
        fig = go.Figure()
        fig.add_trace(go.Scatter3d(x=x[:, 0], y=x[:, 1], z=x[:, 2], mode=mode))
    if x.ndim == 3:
        indices = [ATOM_NUMBERING[a] for a in atoms]
        fig = go.Figure()
        for idx in indices:
            fig.add_scatter3d(
                x=x[:, idx, 0],
                y=x[:, idx, 1],
                z=x[:, idx, 2],
                mode=mode,
                name=atoms[idx],
                text=residue_ids,
            )
    return fig



"""Functions for working with RNA Secondary Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Emmanuele Rossi, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
# This submodule is heavily inspired by: https://github.com/emalgorithm/rna-design/blob/aec77a18abe4850958d6736ec185a6f8cbfdf20c/src/util.py#L9

import os
from typing import Callable, List, Optional, Union

import networkx as nx
from loguru import logger as log

import graphein.protein.graphs as gp
from graphein.rna.config import BpRNAConfig, RNAGraphConfig
from graphein.rna.constants import (
    RNA_BASE_COLORS,
    RNA_BASES,
    SUPPORTED_DOTBRACKET_NOTATION,
)
from graphein.rna.nussinov import nussinov
from graphein.rna.utils import read_dbn_file
from graphein.utils.utils import (
    annotate_edge_metadata,
    annotate_graph_metadata,
    annotate_node_metadata,
    compute_edges,
)


def validate_rna_sequence(s: str) -> None:
    """
    Validate RNA sequence. This ensures that it only contains supported bases.

    Supported bases are: ``"A", "U", "G", "C", "I"``
    Supported bases can be accessed in
        :const:`~graphein.rna.constants.RNA_BASES`

    :param s: Sequence to validate
    :type s: str
    :raises ValueError: Raises ValueError if the sequence contains an
        unsupported base character
    """
    letters_used = set(s)
    if not letters_used.issubset(RNA_BASES):
        offending_letter = letters_used.difference(RNA_BASES)
        position = s.index(offending_letter)
        raise ValueError(
            f"Invalid letter {offending_letter} found at position {position} \
                in the sequence {s}."
        )


def validate_lengths(db: str, seq: str) -> None:
    """
    Check lengths of dotbracket and sequence match.

    :param db: Dotbracket string to check
    :type db: str
    :param seq: RNA nucleotide sequence to check.
    :type seq: str
    :raises ValueError: Raises ``ValueError`` if lengths of dotbracket and
        sequence do not match.
    """
    if len(db) != len(seq):
        raise ValueError(
            f"Length of dotbracket ({len(db)}) does not match length of \
                sequence ({len(seq)})."
        )


def validate_dotbracket(db: str):
    """
    Sanitize dotbracket string. This ensures that it only has supported symbols.

    See: :const:`~graphein.rna.constants.SUPPORTED_DOTBRACKET_NOTATION`

    :param db: Dotbracket notation string
    :type db: str
    :raises ValueError: Raises ValueError if dotbracket notation contains
        unsupported symbols
    """
    chars_used = set(db)
    if not chars_used.issubset(SUPPORTED_DOTBRACKET_NOTATION):
        offending_letter = chars_used.difference(SUPPORTED_DOTBRACKET_NOTATION)
        position = db.index(offending_letter)
        raise ValueError(
            f"Invalid letter {offending_letter} found at position {position} \
                in the sequence {db}."
        )


def construct_rna_graph_3d(
    config: Optional[RNAGraphConfig] = None,
    name: Optional[str] = None,
    path: Optional[Union[str, os.PathLike]] = None,
    pdb_code: Optional[str] = None,
    chain_selection: str = "all",
    model_index: int = 1,
    rna_df_processing_funcs: Optional[List[Callable]] = None,
    edge_construction_funcs: Optional[List[Callable]] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
) -> nx.Graph:
    """
    Constructs RNA structure graph from a ``pdb_code`` or ``path``.

    Users can provide a :class:`~graphein.rna.config.RNAGraphConfig`
    object to specify construction parameters.

    However, config parameters can be overridden by passing arguments directly
    to the function.

    :param config: :class:`~graphein.rna.config.RNAGraphConfig` object. If
        ``None``, defaults to config in ``graphein.rna.config``.
    :type config: graphein.protein.config.RNAGraphConfig, optional
    :param path: Path to PDB or MMTF to build graph from. Default is
        ``None``.
    :type path: str, optional
    :param pdb_code: 4-character PDB accession pdb_code to build graph from.
        Default is ``None``.
    :type pdb_code: str, optional
    :param chain_selection: String of nucleotide chains to include in graph.
        E.g ``"ABDF"`` or ``"all"``. Default is ``"all"``.
    :type chain_selection: str
    :param model_index: Index of model to use in the case of structural
        ensembles. Default is ``1``.
    :type model_index: int
    :param df_processing_funcs: List of DataFrame processing functions.
        Default is ``None``.
    :type df_processing_funcs: List[Callable], optional
    :param edge_construction_funcs: List of edge construction functions.
        Default is ``None``.
    :type edge_construction_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of edge annotation functions.
        Default is ``None``.
    :type edge_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of node annotation functions.
        Default is ``None``.
    :type node_annotation_funcs: List[Callable], optional
    :param graph_annotation_funcs: List of graph annotation function.
        Default is ``None``.
    :type graph_annotation_funcs: List[Callable]
    :return: RNA Structure Graph
    :type: nx.Graph
    """

    # If no config is provided, use default
    if config is None:
        config = RNAGraphConfig()

    # Get name from pdb_file is no pdb_code is provided
    if path and (pdb_code is None):
        pdb_code = gp.get_protein_name_from_filename(path)

    # If config params are provided, overwrite them
    config.rna_df_processing_functions = (
        rna_df_processing_funcs
        if config.rna_df_processing_functions is None
        else config.rna_df_processing_functions
    )
    config.edge_construction_functions = (
        edge_construction_funcs
        if config.edge_construction_functions is None
        else config.edge_construction_functions
    )
    config.node_metadata_functions = (
        node_annotation_funcs
        if config.node_metadata_functions is None
        else config.node_metadata_functions
    )
    config.graph_metadata_functions = (
        graph_annotation_funcs
        if config.graph_metadata_functions is None
        else config.graph_metadata_functions
    )
    config.edge_metadata_functions = (
        edge_annotation_funcs
        if config.edge_metadata_functions is None
        else config.edge_metadata_functions
    )

    raw_df = gp.read_pdb_to_dataframe(
        path,
        pdb_code,
        model_index=model_index,
    )

    raw_df = gp.sort_dataframe(raw_df)
    protein_df = gp.process_dataframe(
        raw_df,
        chain_selection=chain_selection,
        granularity=config.granularity,
        insertions=config.insertions,
        keep_hets=config.keep_hets,
    )

    # Initialise graph with metadata
    g = gp.initialise_graph_with_metadata(
        protein_df=protein_df,
        raw_pdb_df=raw_df,
        name=name,
        pdb_code=pdb_code,
        path=path,
        granularity=config.granularity,
    )
    # Add nodes to graph
    g = gp.add_nodes_to_graph(g)

    # Add config to graph
    g.graph["config"] = config

    # Annotate additional node metadata
    if config.node_metadata_functions is not None:
        g = annotate_node_metadata(g, config.node_metadata_functions)

    # Compute graph edges
    g = compute_edges(
        g,
        funcs=config.edge_construction_functions,
    )

    # Annotate additional graph metadata
    if config.graph_metadata_functions is not None:
        g = annotate_graph_metadata(g, config.graph_metadata_functions)

    # Annotate additional edge metadata
    if config.edge_metadata_functions is not None:
        g = annotate_edge_metadata(g, config.edge_metadata_functions)

    return g


def construct_rna_graph_2d(
    dotbracket: Optional[str],
    sequence: Optional[str],
    name: Optional[str] = None,
    bprna_id: Optional[str] = None,
    edge_construction_funcs: Optional[List[Callable]] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
    use_nussinov: bool = False,
    config: Optional[RNAGraphConfig] = None,
) -> nx.Graph:
    """
    Constructs an RNA secondary structure graph from dotbracket notation.

    :param dotbracket: Dotbracket notation representation of secondary
        structure.
    :type dotbracket: str, optional
    :param sequence: Corresponding sequence RNA bases
    :type sequence: str, optional
    :param bprna_id: bp RNA ID of the RNA secondary structure from which to
        construct the graph. Defaults to ``None``.
    :type bprna_id: str, optional
    :param edge_construction_funcs: List of edge construction functions.
        Defaults to ``None``.
    :type edge_construction_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of edge metadata annotation functions.
        Defaults to ``None``.
    :type edge_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of node metadata annotation functions.
        Defaults to ``None``.
    :type node_annotation_funcs: List[Callable], optional
    :param graph_annotation_funcs: List of graph metadata annotation functions.
        Defaults to ``None``.
    :type graph_annotation_funcs: List[Callable], optional
    :param config: BpRNA Configuration object. Defaults to ``None``.
        Unused unless using a bpRNA to compute a graph.
    :type config: BpRNAConfig, optional
    :return: nx.Graph of RNA secondary structure
    :rtype: nx.Graph
    """
    G = nx.Graph(name=name)

    if bprna_id is not None:
        if config is None:
            config = BpRNAConfig()
        sequence, dotbracket = read_dbn_file(config.path)

    # Build node IDs first.
    node_ids = (
        list(range(len(sequence)))
        if sequence
        else list(range(len(dotbracket)))
    )

    if use_nussinov:
        dotbracket = nussinov(sequence)

    # Check sequence and dotbracket lengths match
    if dotbracket and sequence:
        validate_lengths(dotbracket, sequence)

    # add nodes
    G.add_nodes_from(node_ids)
    log.debug(f"Added {len(node_ids)} nodes")

    # Add dotbracket symbol if dotbracket is provided
    if dotbracket:
        validate_dotbracket(dotbracket)
        G.graph["dotbracket"] = dotbracket

        nx.set_node_attributes(
            G,
            dict(zip(node_ids, dotbracket)),
            "dotbracket_symbol",
        )

    # Add nucleotide base info if sequence is provided
    if sequence:
        validate_rna_sequence(sequence)
        G.graph["sequence"] = sequence
        nx.set_node_attributes(G, dict(zip(node_ids, sequence)), "nucleotide")
        colors = [RNA_BASE_COLORS[i] for i in sequence]
        nx.set_node_attributes(G, dict(zip(node_ids, colors)), "color")

    # Annotate additional graph metadata
    if graph_annotation_funcs is not None:
        G = annotate_graph_metadata(G, graph_annotation_funcs)

    # Annotate additional node metadata
    if node_annotation_funcs is not None:
        G = annotate_node_metadata(G, node_annotation_funcs)

    # Add edges
    G = compute_edges(G, edge_construction_funcs)

    # Annotate additional edge metadata
    if edge_annotation_funcs is not None:
        G = annotate_edge_metadata(G, edge_annotation_funcs)

    return G


def construct_graph(
    dotbracket: Optional[str] = None,
    sequence: Optional[str] = None,
    bprna_id: Optional[str] = None,
    use_nussinov: bool = False,
    name: Optional[str] = None,
    config: Optional[RNAGraphConfig] = None,
    path: Optional[str] = None,
    pdb_code: Optional[str] = None,
    chain_selection: str = "all",
    rna_df_processing_funcs: Optional[List[Callable]] = None,
    edge_construction_funcs: Optional[List[Callable]] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
) -> nx.Graph:
    # TODO Docstring

    if path is not None or pdb_code is not None:
        return construct_rna_graph_3d(
            config=config,
            name=name,
            path=path,
            pdb_code=pdb_code,
            chain_selection=chain_selection,
            rna_df_processing_funcs=rna_df_processing_funcs,
            edge_construction_funcs=edge_construction_funcs,
            edge_annotation_funcs=edge_annotation_funcs,
            node_annotation_funcs=node_annotation_funcs,
            graph_annotation_funcs=graph_annotation_funcs,
        )
    elif (
        dotbracket is not None or sequence is not None or bprna_id is not None
    ):
        return construct_rna_graph_2d(
            dotbracket=dotbracket,
            sequence=sequence,
            name=name,
            bprna_id=bprna_id,
            use_nussinov=use_nussinov,
            edge_construction_funcs=edge_construction_funcs,
            edge_annotation_funcs=edge_annotation_funcs,
            node_annotation_funcs=node_annotation_funcs,
            graph_annotation_funcs=graph_annotation_funcs,
        )



"""Constants for working with RNA Secondary Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Dict, List

RNA_BASES: List[str] = ["A", "U", "G", "C", "I"]
"""List of allowable RNA Bases."""

RNA_BASE_COLORS: Dict[str, str] = {
    "A": "r",
    "U": "b",
    "G": "g",
    "C": "y",
    "I": "m",
}
"""Maps RNA bases (:const:`~graphein.rna.constants.RNA_BASES`) to a colour for visualisations."""

CANONICAL_BASE_PAIRINGS: Dict[str, List[str]] = {
    "A": ["U"],
    "U": ["A"],
    "G": ["C"],
    "C": ["G"],
}
"""Maps standard RNA bases to their canonical base pairings."""

WOBBLE_BASE_PAIRINGS: Dict[str, List[str]] = {
    "A": ["I"],
    "U": ["G", "I"],
    "G": ["U"],
    "C": ["I"],
    "I": ["A", "C", "U"],
}
"""
Maps RNA bases (:const:`~graphein.rna.constants.RNA_BASES`) to their wobble base pairings.
"""

VALID_BASE_PAIRINGS = {
    key: CANONICAL_BASE_PAIRINGS.get(key, [])
    + WOBBLE_BASE_PAIRINGS.get(key, [])
    for key in set(
        list(CANONICAL_BASE_PAIRINGS.keys())
        + list(WOBBLE_BASE_PAIRINGS.keys())
    )
}
"""
Mapping of RNA bases (:const:`~graphein.rna.constants.RNA_BASES`) to their allowable pairings.
Amalgam of :const:`~graphein.rna.constants.CANONICAL_BASE_PAIRINGS` and :const:`~graphein.rna.constants.WOBBLE_BASE_PAIRINGS`.
"""

SIMPLE_DOTBRACKET_NOTATION: List[str] = ["(", ".", ")"]
"""List of characters in simplest dotbracket notation."""

PSEUDOKNOT_OPENING_SYMBOLS: List[str] = ["[", "{", "<"]
"""List of symbols denoting a pseudoknot opening."""

PSEUDOKNOT_CLOSING_SYMBOLS: List[str] = ["]", "}", ">"]
"""List of symbols denoting a pseudoknot closing."""

SUPPORTED_PSEUDOKNOT_NOTATION: List[str] = (
    PSEUDOKNOT_OPENING_SYMBOLS + PSEUDOKNOT_CLOSING_SYMBOLS
)
"""
List of characters denoting pseudoknots in dotbracket notation.
Amalgam of :const:`~graphein.rna.constants.PSEUDOKNOT_OPENING_SYMBOLS` and :const:`~graphein.rna.constants.PSEUDOKNOT_CLOSING_SYMBOLS`.
"""

SUPPORTED_DOTBRACKET_NOTATION = (
    SIMPLE_DOTBRACKET_NOTATION + SUPPORTED_PSEUDOKNOT_NOTATION
)
"""
List of all valid dotbracket symbols.
Amalgamation of :const:`~graphein.rna.constants.SIMPLE_DOTBRACKET_NOTATION` and :const:`~graphein.rna.constants.SUPPORTED_PSEUDOKNOT_NOTATION`.
"""


SS_BOND_TYPES: List[str] = [
    "phosphodiester_bond",
    "base_pairing",
    "pseudoknot",
]
"""List of valid secondary structure bond types."""

RNA_ATOMS: List[str] = [
    "C1'",
    "C2",
    "C2'",
    "C3'",
    "C4",
    "C4'",
    "C5",
    "C5'",
    "C6",
    "C8",
    "N1",
    "N2",
    "N3",
    "N4",
    "N6",
    "N7",
    "N9",
    "O2",
    "O2'",
    "O3'",
    "O4",
    "O4'",
    "O5'",
    "O6",
    "OP1",
    "OP2",
    "P",
]
"""List of valid RNA atoms found in PDB structures."""

PHOSPHORIC_ACID_ATOMS: List[str] = ["P", "OP1", "OP2"]

RIBOSE_ATOMS: List[str] = [
    "C1'",
    "O1'",
    "C2'",
    "O2'",
    "C3'",
    "O3'",
    "C4'",
    "O4'",
    "C5'",
    "O5'",
]

RNA_BACKBONE_ATOMS = RIBOSE_ATOMS + PHOSPHORIC_ACID_ATOMS

RNA_ATOMIC_RADII: Dict[str, float] = {
    "Csb": 0.77,
    "Cres": 0.72,
    "Cdb": 0.67,
    "Nsb": 0.70,
    "Ndb": 0.62,
    "H": 0.37,
    "Osb": 0.67,
    "Odb": 0.60,
    "P": 0.92,
}
"""List of atomic radii (in angstroms) from Heyrovska 2008."""

PHOSPHORIC_ACID_BOND_STATE: Dict[str, str] = {
    "P": "P",
    "OP1": "Osb",  # TODO check this? Maybe incorrect
    "OP2": "Odb",  # TODO check this? Maybe incorrect
}

RIBOSE_BOND_STATE: Dict[str, str] = {
    "C1'": "Cres",
    "O1'": "Osb",
    "C2'": "Cres",
    "O2'": "Osb",
    "C3'": "Cres",
    "O3'": "Osb",
    "C4'": "Cres",
    "O4'": "Osb",
    "C5'": "Cres",
    "O5'": "Osb",
    "H": "H",
}

RNA_BACKBONE_BOND_STATE: Dict[str, str] = {
    **RIBOSE_BOND_STATE,
    **PHOSPHORIC_ACID_BOND_STATE,
}

DEOXYRIBOSE_BOND_STATE: Dict[str, str] = {
    "C1'": "Cres",
    "O1'": "Osb",
    "C2'": "Cres",
    "O2'": "Osb",
    "C3'": "Cres",
    "O3'": "Osb",
    "C4'": "Cres",
    "O4'": "Osb",
    "C5'": "Cres",
    "O5'": "Osb",
    "H": "H",
}

A_ATOMS: List[str] = [
    "N1",
    "C2",
    "N3",
    "C4",
    "C5",
    "C6",
    "N6",
    "N7",
    "C8",
    "N9",
]

C_ATOMS: List[str] = [
    "N1",
    "C2",
    "O2",
    "N3",
    "C4",
    "N4",
    "C5",
    "C6",
]

G_ATOMS: List[str] = [
    "N1",
    "C2",
    "N2",
    "N3",
    "C4",
    "C5",
    "C6",
    "O6",
    "N7",
    "C8",
    "N9",
]

T_ATOMS: List[str] = []

RNA_ATOM_BOND_STATE: Dict[str, Dict[str, str]] = {
    "A": {
        "N1": "Ndb",
        "C2": "Cres",
        "N3": "Ndb",
        "C4": "Cres",
        "C5": "Cdb",
        "C6": "Cres",
        "N6": "Nsb",
        "N7": "Ndb",
        "C8": "Cdb",
        "N9": "Nsb",
        "H": "H",
        **RNA_BACKBONE_BOND_STATE,
    },
    "C": {
        "N1": "Nsb",
        "C2": "Cdb",
        "O2": "Odb",
        "N3": "Ndb",
        "C4": "Cdb",
        "N4": "Nsb",
        "C5": "Cres",
        "C6": "Cdb",
        "H": "H",
        **RNA_BACKBONE_BOND_STATE,
    },
    "G": {
        "N1": "Nsb",
        "C2": "Cdb",
        "N2": "Nsb",
        "N3": "Ndb",
        "C4": "Cdb",
        "C5": "Cres",
        "C6": "Cdb",
        "O6": "Odb",
        "N7": "Ndb",
        "C8": "Cdb",
        "N9": "Nsb",
        "H": "H",
        **RNA_BACKBONE_BOND_STATE,
    },
    "T": {
        "N1": "Nsb",
        "C2": "Cdb",
        "O2": "Odb",
        "N3": "Nsb",
        "C4": "Cdb",
        "O4": "Odb",
        "C5": "Cres",
        "C": "Csb",  # TODO check this? Maybe incorrect
        "C6": "Cdb",
        "H": "H",
        **RNA_BACKBONE_BOND_STATE,
    },
    "U": {
        "N1": "Nsb",
        "C2": "Cdb",
        "O2": "Odb",
        "N3": "Nsb",
        "C4": "Cdb",
        "O4": "Odb",
        "C5": "Cres",
        "C6": "Cdb",
        "H": "H",
        **RNA_BACKBONE_BOND_STATE,
    },
}
"""List of atoms mapped to bondstates from Heyrovska 2008."""



"""Pytorch Geometric Dataset classes for Protein Graphs."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
from collections.abc import Sequence
from difflib import get_close_matches
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional, Union

import networkx as nx
import numpy as np
from loguru import logger as log
from tqdm import tqdm

from graphein.protein.utils import (  # download_pdb,
    download_alphafold_structure,
    download_pdb_multiprocessing,
    read_fasta,
)
from graphein.utils.dependencies import import_message

from ..folding_utils import esmfold
from .data import Protein, ProteinBatch, to_protein_mp

try:
    import torch
    from torch_geometric.data import Data, Dataset, InMemoryDataset
except ImportError:
    import_message(
        "graphein.ml.datasets.torch_geometric_dataset",
        "torch_geometric",
        conda_channel="pyg",
        pip_install=True,
    )

IndexType = Union[slice, torch.Tensor, np.ndarray, Sequence]


class ProteinDataset(Dataset):
    def __init__(
        self,
        root: str,
        pdb_dir: str,
        out_dir: str,
        overwrite: bool = False,
        paths: Optional[List[str]] = None,
        pdb_codes: Optional[List[str]] = None,
        uniprot_ids: Optional[List[str]] = None,
        sequences: Optional[Union[List[str], str]] = None,
        graph_labels: Optional[List[torch.Tensor]] = None,
        node_labels: Optional[List[torch.Tensor]] = None,
        chain_selections: Optional[List[str]] = None,
        graph_transformation_funcs: Optional[List[Callable]] = None,
        pdb_transform: Optional[List[Callable]] = None,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        pre_filter: Optional[Callable] = None,
        num_cores: int = 16,
        af_version: int = 2,
    ):
        """Dataset class for protein graphs.

        Dataset base class for creating graph datasets.
        See `here <https://pytorch-geometric.readthedocs.io/en/latest/notes/
        create_dataset.html>`__ for the accompanying tutorial.

        :param root: Root directory where the dataset should be saved.
        :type root: str
        :param pdb_dir: Directory to save PDB files to. If not specified
            (`pdb_dir=`None``), will default to ``root/pdb``. Defaults to
            ``None``.
        :type pdb_dir: str, optional
        :param out_dir: Directory to save output files to. If not specified
            (``out_dir=None``), will default to ``root/processed``. Defaults to
            ``None``.
        :param overwrite: Whether to overwrite existing files. Defaults to
            ``False``.
        :type overwrite: bool, optional
        :param paths: List of full path of PDB or MMTF files to load. Defaults to
            ``None``.
        :type paths: Optional[List[str]], optional
        :param pdb_codes: List of PDB codes to download and parse from the PDB.
            Defaults to ``None``.
        :type pdb_codes: Optional[List[str]], optional
        :param uniprot_ids: List of Uniprot IDs to download and parse from
            Alphafold Database. Defaults to ``None``.
        :type uniprot_ids: Optional[List[str]], optional
        :param sequences: List of protein sequences to predict structures of with
            ESMFold. Also accepts a path to a fasta file. Defaults to ``None``.
        :type sequences: Optional[Union[List[str], str]], optional
        :param graph_label_map: Dictionary mapping PDB/Uniprot IDs to
            graph-level labels. Defaults to ``None``.
        :type graph_label_map: Optional[Dict[str, Tensor]], optional
        :param node_label_map: Dictionary mapping PDB/Uniprot IDs to node-level
            labels. Defaults to ``None``.
        :type node_label_map: Optional[Dict[str, torch.Tensor]], optional
        :param chain_selection_map: Dictionary mapping, defaults to ``None``.
        :type chain_selection_map: Optional[Dict[str, List[str]]], optional
        :param graph_transformation_funcs: List of functions that consume a
            ``nx.Graph`` and return a ``nx.Graph``. Applied to graphs after
            construction but before conversion to pyg. Defaults to ``None``.
        :type graph_transformation_funcs: Optional[List[Callable]], optional
        :param pdb_transform: List of functions that consume a list of paths to
            the downloaded structures. This provides an entry point to apply
            pre-processing from bioinformatics tools of your choosing. Defaults
            to ``None``.
        :type pdb_transform: Optional[List[Callable]], optional
        :param transform: A function/transform that takes in a
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before every access.
            Defaults to ``None``.
        :type transform: Optional[Callable], optional
        :param pre_transform:  A function/transform that takes in an
            ``torch_geometric.data.Data`` object and returns a transformed
            version. The data object will be transformed before being saved to
            disk. Defaults to ``None``.
        :type pre_transform: Optional[Callable], optional
        :param pre_filter:  A function that takes in a
            ``torch_geometric.data.Data`` object and returns a boolean value,
            indicating whether the data object should be included in the final
            dataset. Optional, defaults to ``None``.
        :type pre_filter: Optional[Callable], optional
        :param num_cores: Number of cores to use for multiprocessing of graph
            construction, defaults to ``16``.
        :type num_cores: int, optional
        :param af_version: Version of AlphaFoldDB structures to use,
            defaults to ``2``.
        :type af_version: int, optional
        """
        # Setup paths
        self.root = root
        self.pdb_dir = pdb_dir
        self.out_dir = out_dir
        self.overwrite = overwrite

        # Setup inputs
        self.pdb_codes = (
            [pdb.lower() for pdb in pdb_codes]
            if pdb_codes is not None
            else None
        )
        self.uniprot_ids = (
            [up.upper() for up in uniprot_ids]
            if uniprot_ids is not None
            else None
        )
        if sequences is not None:
            self.sequences = sequences
            self.fold_sequences()

        self.paths = paths
        if self.paths is None:
            if self.pdb_codes and self.uniprot_ids:
                self.structures = self.pdb_codes + self.uniprot_ids
            elif self.pdb_codes:
                self.structures = pdb_codes
            elif self.uniprot_ids:
                self.structures = uniprot_ids
        # Use local saved pdb_files instead of download or move them to
        # self.root/raw dir
        else:
            if isinstance(self.paths, list):
                self.structures = [
                    os.path.splitext(os.path.split(path)[-1])[0]
                    for path in self.paths
                ]
                self.path, _ = os.path.split(self.paths[0])

        # Labels & chains
        self.chain_selections = chain_selections
        if self.chain_selections is not None:
            self.chain_selections = [
                chain if chain is not None else "all"
                for chain in self.chain_selections
            ]
        self.graph_labels = graph_labels
        self.node_labels = node_labels
        self.validate_input()
        self.bad_pdbs: List[str] = []

        # Configs
        self.num_cores = num_cores
        self.pdb_transform = pdb_transform
        self.graph_transformation_funcs = graph_transformation_funcs
        self.af_version = af_version

        log.info(f"Creating a dataset from {len(self.structures)} structures.")
        log.info(f"PDB Directory: {self.raw_dir}")
        log.info(f"Output Directory: {self.processed_dir}")

        # Do overwriting
        if self.overwrite:
            log.info(
                "Overwrite flag is set. Existing files will be overwritten."
            )
        else:
            log.info(
                "Overwrite flag is not set. Existing files will not be regenerated."
            )
        if self.overwrite:
            log.info("Removing existing (processed) files...")
            for f in tqdm(self.processed_file_names):
                if os.path.exists(os.path.join(self.processed_dir, f)):
                    os.remove(os.path.join(self.processed_dir, f))
        super().__init__(
            root,
            transform=transform,
            pre_transform=pre_transform,
            pre_filter=pre_filter,
        )

    @property
    def raw_file_names(self) -> List[str]:
        """Names of raw files in the dataset."""
        return [f"{pdb}.pdb" for pdb in self.structures]

    @property
    def processed_file_names(self) -> List[str]:
        """Names of processed files to look for"""
        if self.chain_selections is not None:
            return [
                f"{pdb}_{chain}.pt"
                for pdb, chain in zip(self.structures, self.chain_selections)
            ]
        else:
            return [f"{pdb}.pt" for pdb in self.structures]

    @property
    def raw_dir(self) -> str:
        if hasattr(self, "pdb_dir"):
            return self.pdb_dir
        else:
            return os.path.join(self.root, "raw")

    @property
    def processed_dir(self) -> str:
        if hasattr(self, "out_dir"):
            return self.out_dir
        else:
            return os.path.join(self.root, "processed")

    def validate_input(self):
        if self.graph_labels is not None:
            assert len(self.structures) == len(
                self.graph_labels
            ), "Number of proteins and graph labels must match"
        if self.node_labels is not None:
            assert len(self.structures) == len(
                self.node_labels
            ), "Number of proteins and node labels must match"
        if self.chain_selections is not None:
            assert len(self.structures) == len(
                self.chain_selections
            ), "Number of proteins and chain selections must match"
            assert len(
                {
                    f"{pdb}_{chain}"
                    for pdb, chain in zip(
                        self.structures, self.chain_selections
                    )
                }
            ) == len(self.structures), "Duplicate protein/chain combinations"

    def download(self):
        """Download the PDB files from RCSB or AlphaFold."""
        if self.pdb_codes:
            # Only download undownloaded PDBs
            to_download = [
                pdb
                for pdb in set(self.pdb_codes)
                if (
                    not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
                    or self.overwrite == True
                )
            ]
            log.info(f"Downloading {len(to_download)} PDBs")
            download_pdb_multiprocessing(
                to_download,
                self.raw_dir,
                max_workers=self.num_cores,
                strict=False,
            )
            self.bad_pdbs = self.bad_pdbs + [
                pdb
                for pdb in set(self.pdb_codes)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
            ]
            if len(self.bad_pdbs) > 0:
                log.warning(
                    f"Failed to download {len(self.bad_pdbs)} PDBs: {self.bad_pdbs}"
                )
            log.info("PDB download complete")
        if self.uniprot_ids:
            log.info(
                f"Downloading {len(self.uniprot_ids)}AlphaFold structures"
            )
            to_download = [
                pdb
                for pdb in set(self.uniprot_ids)
                if not os.path.exists(Path(self.raw_dir) / f"{pdb}.pdb")
                or self.overwrite == True
            ]
            [
                download_alphafold_structure(
                    uniprot,
                    out_dir=self.raw_dir,
                    version=self.af_version,
                    aligned_score=False,
                )
                for uniprot in tqdm(to_download)
            ]
            log.info("AlphaFold download complete")

    def fold_sequences(self):
        if isinstance(self.sequences, list):
            self.sequences = dict(enumerate(self.sequences))
        elif isinstance(self.sequences, str):
            assert os.path.exists(
                self.sequences
            ), f"Sequence file {self.sequences} does not exist"
            self.sequences = read_fasta(self.sequences)
        else:
            raise ValueError(
                "Sequences must be a list of sequences or a path to a file containing sequences"
            )

        log.info(f"Folding {len(self.sequences)} sequences...")
        if not hasattr(self, "structures"):
            self.structures = []
        for id, seq in tqdm(self.sequences.items()):
            # Add ID to list of structures
            self.structures += [id]
            # Check for existence of file
            if (
                os.path.exists(os.path.join(self.raw_dir, f"{id}.pdb"))
                and not self.overwrite
            ):
                continue
            else:
                esmfold(seq, out_path=os.path.join(self.raw_dir, f"{id}.pdb"))

    def len(self) -> int:
        """Returns length of data set (number of structures)."""
        return len(self.structures)

    def transform_pdbs(self):
        """
        Performs pre-processing of PDB structures before constructing graphs.
        """
        structure_files = [
            f"{self.raw_dir}/{pdb}.pdb" for pdb in self.structures
        ]
        for func in self.pdb_transform:
            func(structure_files)

    def transform_protein_graphs(self, graph: Protein) -> Protein:
        """
        Applies a series of transformations to a protein graph.

        :param graph: Protein graph
        :type graph: Protein
        :return: Transformed protein graph
        :rtype: Protein
        """
        for func in self.graph_transformation_funcs:
            graph = func(graph)
        return graph

    def process(self):
        """Processes structures from files into PyTorch Geometric Data."""
        log.info("Processing PDB files into graphs...")
        # Preprocess PDB files
        if self.pdb_transform:
            log.info(f"Applying PDB transformations: {self.pdb_transform}")
            self.transform_pdbs()
        else:
            log.info("No PDB transformations specified.")

        # Chunk dataset for parallel processing
        chunk_size = 128

        def divide_chunks(l: List[str], n: int = 2) -> Generator:
            for i in range(0, len(l), n):
                yield l[i : i + n]

        examples: List[List[str]] = list(
            divide_chunks(self.structures, chunk_size)
        )
        idxs = list(divide_chunks(range(len(self.structures)), chunk_size))

        if self.chain_selections is None:
            chain_selections = ["all"] * len(self.structures)
        else:
            chain_selections = self.chain_selections

        chain_selections: List[List[str]] = list(
            divide_chunks(chain_selections, chunk_size)
        )

        for idx, pdbs, chains in tqdm(zip(idxs, examples, chain_selections)):
            # Create graph objects
            file_names: List[str] = [f"{self.raw_dir}/{pdb}" for pdb in pdbs]
            file_names: List[str] = [
                f if f.endswith(".pdb") else f"{f}.pdb" for f in file_names
            ]
            print(file_names)
            graphs = to_protein_mp(
                paths=file_names,
                # config=self.config,
                # TODO args
                chain_selections=chains,
                num_cores=self.num_cores,
            )

            if self.graph_transformation_funcs is not None:
                graphs = [self.transform_protein_graphs(g) for g in graphs]

            # Assign labels
            if self.graph_labels is not None:
                for graph_idx, label_idx in enumerate(idx):
                    graphs[graph_idx].graph_y = self.graph_labels[label_idx]
            if self.node_labels is not None:
                for graph_idx, label_idx in enumerate(idx):
                    graphs[graph_idx].node_y = self.node_labels[label_idx]

            # Apply filters and transforms
            if self.pre_filter is not None:
                graphs = [g for g in graphs if self.pre_filter(g)]

            if self.pre_transform is not None:
                graphs = [self.pre_transform(data) for data in graphs]

            # Save graphs as pickles
            for i, (pdb, chain) in enumerate(zip(pdbs, chains)):
                if self.chain_selections is None:
                    torch.save(
                        graphs[i],
                        os.path.join(self.processed_dir, f"{pdb}.pt"),
                    )
                else:
                    torch.save(
                        graphs[i],
                        os.path.join(self.processed_dir, f"{pdb}_{chain}.pt"),
                    )

    def get(self, idx: int) -> Protein:
        """
        Returns Protein Data object for a given index.

        :param idx: Index to retrieve.
        :type idx: int
        :return: Protein data object.
        """
        if isinstance(idx, int):
            if self.chain_selections is None:
                return torch.load(
                    os.path.join(
                        self.processed_dir, f"{self.structures[idx]}.pt"
                    ),
                    weights_only=False,
                )
            return torch.load(
                os.path.join(
                    self.processed_dir,
                    f"{self.structures[idx]}_{self.chain_selections[idx]}.pt",
                ),
                weights_only=False,
            )
        elif isinstance(idx, str):
            try:
                return torch.load(
                    os.path.join(self.processed_dir, f"{idx}.pt"),
                    weights_only=False,
                )
            except FileNotFoundError:
                files = os.listdir(self.processed_dir)
                similar_files = get_close_matches(
                    f"{idx}.pt", files, n=5, cutoff=0.7
                )
                log.error(
                    f"File {idx}.pt not found. Did you mean: {similar_files}"
                )

    def __getitem__(
        self,
        idx: Union[int, np.integer, IndexType, str],
    ) -> Union["Dataset", Data]:
        r"""In case :obj:`idx` is of type integer, will return the data object
        at index :obj:`idx` (and transforms it in case :obj:`transform` is
        present).
        In case :obj:`idx` is a slicing object, *e.g.*, :obj:`[2:5]`, a list, a
        tuple, or a :obj:`torch.Tensor` or :obj:`np.ndarray` of type long or
        bool, will return a subset of the dataset at the specified indices."""
        if (
            isinstance(idx, (int, np.integer))
            or (isinstance(idx, torch.Tensor) and idx.dim() == 0)
            or (isinstance(idx, np.ndarray) and np.isscalar(idx))
        ):
            data = self.get(self.indices()[idx])
            data = data if self.transform is None else self.transform(data)
            return data
        elif isinstance(idx, str):
            data = self.get(idx)
            data = data if self.transform is None else self.transform(data)
            return data
        else:
            return self.index_select(idx)


class ProteinGraphListDataset(InMemoryDataset):
    def __init__(
        self, root: str, data_list: List[Data], name: str, transform=None
    ):
        """Creates a dataset from a list of PyTorch Geometric Data objects.

        :param root: Root directory where the dataset is stored.
        :type root: str
        :param data_list: List of protein graphs as PyTorch Geometric Data
            objects.
        :type data_list: List[Data]
        :param name: Name of dataset. Data will be saved as ``data_{name}.pt``.
        :type name: str
        :param transform: A function/transform that takes in an
            :obj:`torch_geometric.data.Data` object and returns a transformed
            version. The data object will be transformed before every access.
            (default: :obj:`None`)
        :type transform: Optional[Callable], optional
        """
        self.data_list = data_list
        self.name = name
        super().__init__(root, transform)
        self.data, self.slices = torch.load(
            self.processed_paths[0], weights_only=False
        )

    @property
    def processed_file_names(self):
        """The name of the files in the :obj:`self.processed_dir` folder that
        must be present in order to skip processing."""
        return f"{self.name}.pt"

    def process(self):
        """Saves data files to disk."""
        torch.save(self.collate(self.data_list), self.processed_paths[0])
        log.info(f"Saved dataset to disk: {self.name}")



# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Dict, List, Optional, Union

from pydantic import BaseModel


class STRINGConfig(BaseModel):
    """
    Config for specifying parameters for API calls to STRINGdb. Full documentation can be found: https://string-db.org/help/api/

    :param species: NCBI taxon identifiers, defaults to 9606 (human)
    :type species: int, optional
    :param required_score: Threshold of significance to include a interaction, a number between 0 and 1000 (default depends on the network)
    :type required_score: int, optional
    :param network_type: Network type: "functional" (default), "physical"
    :type network_type: str, optional
    :param add_nodes: Adds a number of proteins to the network based on their confidence score, e.g., extends the interaction neighborhood of selected proteins to desired value, defaults to 50
    :type add_nodes: int, optional
    :param show_query_node_labels: When available use submitted names in the preferredName column when (0 or 1) (default:0)
    :type show_query_node_labels: bool, optional
    """

    species: Optional[int] = 9606  # NCBI taxon identifiers
    required_score: Optional[int] = (
        50  # threshold of significance to include a interaction, a number between 0 and 1000 (default depends on the network)
    )
    network_type: Optional[str] = (
        "functional"  # network type: functional (default), physical
    )
    add_nodes: Optional[int] = (
        0  # adds a number of proteins to the network based on their confidence score, e.g., extends the interaction neighborhood of selected proteins to desired value
    )
    show_query_node_labels: Optional[bool] = (
        0  # when available use submitted names in the preferredName column when (0 or 1) (default:0)
    )


class BioGridConfig(BaseModel):
    """
    Config for specifying parameters for API calls to BIOGRID. A full description of the parameters can be found at : https://wiki.thebiogrid.org/doku.php/biogridrest

    :param searchNames: If ‘true’, the interactor OFFICIAL_SYMBOL will be examined for a match with the geneList.
    :type searchNames: bool, optional
    :param max: Number of results to fetch, defaults to 10,000
    :type max: int, optional
    :param interSpeciesExcluded: If ‘true’, interactions with interactors from different species will be excluded, defaults to True
    :type interSpeciesExcluded: bool, optional
    :param selfInteractionsExcluded:  If ‘true’, interactions with one interactor will be excluded, defaults to False
    :type selfInteractionsExcluded: bool, optional.
    :param evidenceList: Any interaction evidence with its Experimental System in the list will be excluded from the results unless includeEvidence is set to true., defaults to "" (empty string)
    :type evidenceList: str, optional
    :param includeEvidence: If set to true, any interaction evidence with its Experimental System in the evidenceList will be included in the result, defaults to False
    :type includeEvidence: bool, optional
    :param searchIDs: If ‘true’, the interactor ENTREZ_GENE, ORDERED LOCUS and SYSTEMATIC_NAME (orf) will be examined for a match with the geneList. Defaults to True
    :type searchIDs: bool, optional
    :param searchNames: # If ‘true’, the interactor OFFICIAL_SYMBOL will be examined for a match with the geneList. Defaults to True.
    :type searchNames: bool, optional
    :param searchSynonyms:  If ‘true’, the interactor SYNONYMS will be examined for a match with the geneList. Defaults to True.
    :type searchSynonyms: bool, optional
    :param searchBiogridIds:  If ‘true’, the entries in 'GENELIST' will be compared to BIOGRID internal IDS which are provided in all Tab2 formatted files. Defaults to True
    :type seachBiogridIds: bool, optional
    :param additionalIdentifierTypes: Identifier types on this list are examined for a match with the geneList. Defaults to ""
    :type additionalIdentifierTypes: str, optional
    :param excludeGenes: If ‘true’, interactions containing genes in the geneList will be excluded from the results. Defaults to False
    :type excludeGenes: bool, optional
    :param includeInteractors:  If ‘true’, in addition to interactions between genes on the geneList, interactions will also be fetched which have only one interactor on the geneList. Defaults to True
    :type includeInteractors: bool, optional
    :param includeInteractorInteractions: # If ‘true’ interactions between the geneList’s first order interactors will be included. Defaults to False
    :type includeInteractorInteractions: bool, optional
    :param pubmedList: Interactions will be fetched whose Pubmed Id is/ is not in this list, depending on the value of excludePubmeds. Defaults to ""
    :type pubmedList: str, optional
    :param excludePubmeds: If ‘false’, interactions with Pubmed ID in pubmedList will be included in the results; if ‘true’ they will be excluded. Defaults to False
    :type excludePubmeds: bool, optional
    :param htpThreshold: Interactions whose Pubmed ID has more than this number of interactions will be excluded from the results. Ignored if excludePubmeds is ‘false’. Defaults to 20.
    :type htpThreshold: int, optional
    :param throughputTag: If set to 'low or 'high', only interactions with 'Low throughput' or 'High throughput' in the 'throughput' field will be returned. Defaults to "any"
    :type throughputTag: str, optional
    """

    searchNames: Optional[bool] = (
        True  # If ‘true’, the interactor OFFICIAL_SYMBOL will be examined for a match with the geneList.
    )
    max: Optional[int] = 10000  # Number of results to fetch
    interSpeciesExcluded: Optional[bool] = (
        True  # If ‘true’, interactions with interactors from different species will be excluded.
    )
    selfInteractionsExcluded: Optional[bool] = (
        False  # If ‘true’, interactions with one interactor will be excluded.
    )
    evidenceList: Optional[str] = (
        ""  # Any interaction evidence with its Experimental System in the list will be excluded from the results unless includeEvidence is set to true.
    )
    includeEvidence: Optional[bool] = (
        False  # If set to true, any interaction evidence with its Experimental System in the evidenceList will be included in the result
    )
    searchIds: Optional[bool] = (
        True  # If ‘true’, the interactor ENTREZ_GENE, ORDERED LOCUS and SYSTEMATIC_NAME (orf) will be examined for a match with the geneList.
    )
    searchNames: Optional[bool] = (
        True  # If ‘true’, the interactor OFFICIAL_SYMBOL will be examined for a match with the geneList.
    )
    searchSynonyms: Optional[bool] = (
        True  # If ‘true’, the interactor SYNONYMS will be examined for a match with the geneList.
    )
    searchBiogridIds: Optional[bool] = (
        True  # If ‘true’, the entries in 'GENELIST' will be compared to BIOGRID internal IDS which are provided in all Tab2 formatted files.
    )
    additionalIdentifierTypes: Optional[str] = (
        ""  # Identifier types on this list are examined for a match with the geneList.
    )
    excludeGenes: Optional[bool] = (
        False  # If ‘true’, interactions containing genes in the geneList will be excluded from the results.
    )
    includeInteractors: Optional[bool] = (
        True  # If ‘true’, in addition to interactions between genes on the geneList, interactions will also be fetched which have only one interactor on the geneList
    )
    includeInteractorInteractions: Optional[bool] = (
        False  # If ‘true’ interactions between the geneList’s first order interactors will be included.
    )
    pubmedList: Optional[str] = (
        ""  # Interactions will be fetched whose Pubmed Id is/ is not in this list, depending on the value of excludePubmeds.
    )
    excludePubmeds: Optional[bool] = (
        False  # If ‘false’, interactions with Pubmed ID in pubmedList will be included in the results; if ‘true’ they will be excluded.
    )
    htpThreshold: Optional[int] = (
        20  # Interactions whose Pubmed ID has more than this number of interactions will be excluded from the results. Ignored if excludePubmeds is ‘false’.
    )
    throughputTag: Optional[str] = (
        "any"  # If set to 'low or 'high', only interactions with 'Low throughput' or 'High throughput' in the 'throughput' field will be returned.
    )


class PPIGraphConfig(BaseModel):
    """
    Config for specifying parameters for PPI Graph Construction

    :param paginate: Controls whether or not to paginate API calls. Useful for large queries. Defaults to True
    :type paginate: bool
    :param ncbi_taxon_id: Defaults to 9606 (human)
    :type ncbi_taxon_id: int
    :param kwargs:
    :type kwargs: Dict[str, Union[str, int, float]], optional
    :param string_config: Config Object holding parameters for STRINGdb API calls. Defaults to None
    :type string_config: graphein.ppi.config.STRINGConfig
    :param biogrid_config: Config Object holding parameters for BioGrid API calls. Defaults to None
    :type biogrid_config: graphein.ppi.config.BioGridConfig, optional
    """

    paginate: bool = True
    ncbi_taxon_id: int = 9606
    kwargs: Optional[Dict[str, Union[str, int, float]]] = {
        "STRING_escore": 0.2,  # Keeps STRING interactions with an experimental score >= 0.2
        "BIOGRID_throughputTag": "high",  # Keeps high throughput BIOGRID interactions
    }
    string_config: Optional[STRINGConfig] = None
    biogrid_config: Optional[BioGridConfig] = None

    class Config:
        arbitrary_types_allowed: bool = True



"""Functions for adding node features from AAIndex."""

from typing import Dict, Tuple

import networkx as nx
from loguru import logger as log

from graphein.utils.dependencies import import_message
from graphein.utils.utils import protein_letters_3to1_all_caps

try:
    from pyaaisc import Aaindex
except ImportError:
    message = import_message(
        submodule="graphein.protein.features.nodes.aaindex",
        package="pyaaisc",
        pip_install=True,
        extras=True,
    )
    log.warning(message)


def fetch_AAIndex(accession: str) -> Tuple[str, Dict[str, float]]:
    """
    Fetches AAindex1 dictionary from an accession code. The dictionary maps one-letter AA codes to float values

    :param accession: Aaindex1 accession code
    :type accession: str
    :return: tuple of record titel(str) and dictionary of AA:value mappings
    :rtype: Tuple[str, Dict[str, float]]
    """
    # Initialise AAindex object and get data
    aaindex = Aaindex()
    record = aaindex.get(accession)

    return record.title, record.index_data


def aaindex1(G: nx.Graph, accession: str) -> nx.Graph:
    """Adds AAIndex1 datavalues for a given accession as node features.

    :param G: nx.Graph protein structure graphein to featurise
    :type G: nx.Graph
    :param accession: AAIndex1 accession code for values to use
    :type accession: str
    :return: Protein Structure graph with AAindex1 node features added
    :rtype: nx.Graph
    """

    title, index_data = fetch_AAIndex(accession)

    # TODO: change to allow for a list of all accession numbers?
    G.graph["aaindex1"] = accession + ": " + title

    if G.graph["config"].granularity == "atom":
        raise NameError(
            "AAIndex features cannot be added to atom granularity graph"
        )

    for n in G.nodes:
        residue = n.split(":")[1]
        residue = protein_letters_3to1_all_caps(residue)

        aaindex = index_data[residue]

        G.nodes[n][f"aaindex1_{accession}"] = aaindex

    return G



"""Utilities for working with protein sequences."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from functools import lru_cache
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
from loguru import logger as log

from graphein.utils.dependencies import import_message

from ..resi_atoms import (
    ATOM_NUMBERING_MODIFIED,
    RESI_THREE_TO_1,
    STANDARD_AMINO_ACIDS,
    STANDARD_RESIDUE_ATOMS,
)
from .types import AtomTensor

try:
    import torch
    import torch.nn.functional as F
except ImportError:
    message = import_message(
        "graphein.protein.tensor.sequence",
        "torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.debug(message)


def get_sequence(
    df: pd.DataFrame,
    chains: Union[str, List[str]] = "all",
    insertions: bool = False,
    list_of_three: bool = False,
    three_to_one_map: Optional[str] = None,
    per_atom: bool = False,
) -> Union[str, List[str]]:
    """
    Retrieves the amino acid sequence from a DataFrame of a protein structure.

    :param df: DataFrame of protein structure.
    :type df: pd.DataFrame
    :param chains: List of chains to retain, defaults to ``None`` (all chains).
    :type chains: Optional[List[str]], optional
    :param insertions: Whether or not to keep insertions, defaults to ``False``
    :type insertions: bool, optional
    :param list_of_three: Whether or not to return a list of three letter codes.
        If ``False``, returns the sequence as a one-letter code string.
        Defaults to ``False``.
    :type list_of_three: bool, optional
    :return: Amino acid sequence; either as list of three-letter codes
        (``["ALA", "GLY", "TRP"...]``; ``list_of_three=True``) or string
        (``AGW..``; ``list_of_three=False``).
    :rtype: Union[str, List[str]]
    """
    # Select chains
    if chains != "all":
        if isinstance(chains, str):
            chains = [chains]
        df = df.loc[df.chain_id.isin(chains)]

    # Assign residues IDs
    if "residue_id" not in df.columns:
        df["residue_id"] = (
            df.chain_id
            + ":"
            + df.residue_name
            + ":"
            + df.residue_number.astype(str)
        )
        if insertions:
            df["residue_id"] = df.residue_id + ":" + df.insertion

    # Get residue from unique IDs
    if per_atom:
        sequence = [res.split(":")[1] for res in df.residue_id]
    else:
        sequence = [res.split(":")[1] for res in df.residue_id.unique()]

    # Convert to one letter code
    if list_of_three:
        # if three_to_one_map is not None:
        # map = RESI_THREE_TO_1
        # sequence = [three_to_one_map[res] for res in sequence]
        return sequence
    else:
        return "".join([RESI_THREE_TO_1[res] for res in sequence])


def get_residue_id(
    df: pd.DataFrame, insertions: bool = True, unique: bool = True
) -> List[str]:
    """
    Returns a list of residue IDs from a DataFrame of a protein structure.

    Residue IDs are of the form: ``[chain_id:residue_name:residue_number]`` or
    ``[chain_id:residue_name:residue_number:insertion_code]` if
    ``insertions=True``

    E.g.

    ``["A:SER:1", "A:GLY:2", ...]`` or ``["A:SER:1:A", "A:GLY:2:", ...]``

    :param df: DataFrame of protein structure to extract residue IDs from.
    :type df: pd.DataFrame
    :param insertions: Whether or not to include insertion codes in the residue
        ID.
    :param unique: Whether or not to return only unique residue IDs. If
        ``False``, it returns a (repeated) ID for each atom in the protein. If
        ``True`` we return the unique set of residue IDs. Default is ``True``.
    :type unique: bool, optional
    :return: List of residue IDs.
    :rtype: List[str]
    """
    if "residue_id" not in df.columns:
        df["residue_id"] = (
            df.chain_id
            + ":"
            + df.residue_name
            + ":"
            + df.residue_number.astype(str)
        )
        if insertions:
            df["residue_id"] = df.residue_id + ":" + df.insertion.astype(str)
    return list(df.residue_id.unique()) if unique else list(df.residue_id)


def residue_type_tensor(
    df: pd.DataFrame,
    vocabulary: List[str] = STANDARD_AMINO_ACIDS,
    three_to_one_mapping: Dict[str, str] = RESI_THREE_TO_1,
    one_hot: bool = False,
    insertions: bool = False,
    dtype: torch.dtype = torch.long,
    device: torch.device = torch.device("cpu"),
    per_atom: bool = False,
) -> torch.Tensor:
    """Returns a tensor of the residue types in a protein structure.

    :param df: DataFrame of protein structure.
    :type df: pd.DataFrame
    :param vocabulary: List of allowable residue types, defaults to
        :ref:`graphein.protein.resi_atoms.STANDARD_AMINO_ACIDS`
    :type vocabulary: List[str], optional
    :param three_to_one_mapping: Mapping from three letter to codes to one
        letter amino acid codes, defaults to
        :ref:`graphein.protein.resi_atoms.RESI_THREE_TO_1`
    :type three_to_one_mapping: Dict[str, str], optional
    :param one_hot: Whether to return a tensor of integers denoting residue
        type or whether to one-hot encode residue types, defaults to ``False``.
    :type one_hot: bool, optional
    :param insertions: Whether or not to include insertions, defaults to
        ``False``.
    :type insertions: bool, optional
    :param dtype: torch.dtype of tensor, defaults to ``torch.long``
    :type dtype: torch.dtype, optional
    :param device: Device to create tensor on, defaults to
        ``torch.device("cpu")``
    :type device: torch.device, optional
    :return: Tensor of residue types.
    :rtype: torch.Tensor
    """
    residues = get_sequence(
        df, insertions=insertions, list_of_three=True, per_atom=per_atom
    )

    # Convert to one letter code
    residues_one = []
    for res in residues:
        res = three_to_one_mapping[res]
        if res in vocabulary:
            residues_one.append(res)
        else:
            residues_one.append("X")

    residues = torch.tensor(
        [vocabulary.index(res) for res in residues_one],
        dtype=dtype,
        device=device,
    )

    if one_hot:
        residues = F.one_hot(residues, num_classes=len(vocabulary))
    return residues


@lru_cache(maxsize=1)
def get_atom_indices(
    invert: bool = False,
) -> Union[Dict[str, Tuple[int, ...]], Dict[Tuple[int, ...], str]]:
    """
    Generates a dictionary mapping residue types to atom indices.

    :param invert: If ``True``, inverts the dictionary so that the keys are the
        atom indices and the values are the residue types, defaults to ``False``
    :type invert: bool, optional
    :return: Dictionary mapping residue types to atom indices or vice versa
    :rtype: Union[Dict[str, Tuple[int, ...]], Dict[Tuple[int, ...], str]]
    """
    index_map = {
        k: tuple(sorted([ATOM_NUMBERING_MODIFIED[i] for i in v]))
        for k, v in STANDARD_RESIDUE_ATOMS.items()
    }
    if invert:
        index_map = {v: k for k, v in index_map.items()}
    return index_map


def infer_residue_types(
    x: AtomTensor, fill_value: float = 1e-5, return_sequence: bool = True
) -> Union[str, List[str]]:
    """
    Infers residue types from atom tensor based on non-filled residues.

    .. note:: This function is not robust to structures with missing atoms.

    :param x: Tensor of shape ``(N, Num Atoms, 3)`` where ``N`` is the number
        of residues in the protein, ``Num Atoms`` is the number of atoms
        selected (default is ``37``,
        see: :ref:`graphein.protein.resi_atoms.ATOM_NUMBERING`)
    :type x: AtomTensor
    :param fill_value: Fill value used to denote the absence of an atom in ``x``
        , defaults to ``1e-5``.
    :type fill_value: float, optional
    :param return_sequence: If ``True``, returns the sequence. If ``False``,
        returns a list of three-letter residue codes, defaults to ``True``.
    :type return_sequence: bool, optional
    :return: Sequence or list of three-letter residue codes
    :rtype: Union[str, List[str]]
    """
    rmap: Dict[Tuple[int, ...], str] = get_atom_indices(invert=True)

    def _get_index(y: torch.Tensor) -> str:
        indices = (
            torch.nonzero(torch.sum(y != fill_value, dim=1))
            .squeeze(1)
            .tolist()
        )
        # remove oxt if present
        oxt_index = ATOM_NUMBERING_MODIFIED["OXT"]
        indices = tuple(i for i in indices if i != oxt_index)
        try:
            identity = rmap[indices]
        except KeyError:
            identity = "UNK"
        return identity

    seq = [_get_index(x[i, :, :]) for i in range(x.shape[0])]

    if return_sequence:
        seq = "".join([RESI_THREE_TO_1[res] for res in seq])

    return seq



"""Functions for retrieving molecular data from ChEMBL."""

from typing import Any, Dict

import networkx as nx

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from bioservices import ChEMBL

from graphein.utils.dependencies import requires_python_libs


@requires_python_libs("bioservices")
def get_smiles_from_chembl(chembl_id: str) -> str:
    """Retrieves a SMILE string from a ChEMBL ID.

    :param chembl_id: ChEMBL ID, e.g., ``'CHEMBL1234'``
    :type smiles: str
    :returns: A SMILE string, e.g.,
        ``'COc1cccc(c1)NC(=O)c2cccnc2'``
    :rtype: str
    """
    chembl = ChEMBL()
    data = chembl.get_molecule(chembl_id)
    return data["molecule_structures"]["canonical_smiles"]


@requires_python_libs("bioservices")
def get_chembl_id_from_smiles(smiles: str) -> str:
    """Retrieves a ChEMBL ID from a SMILE string.

    :param smiles: A valid SMILE string, e.g.,
        ``'COc1cccc(c1)NC(=O)c2cccnc2'``
    :type smiles: str
    :return: ChEMBL ID, e.g., ``'CHEMBL1234'``
    :rtype: str
    """
    chembl = ChEMBL()
    data = chembl.get_molecule(smiles)
    return data["molecule_chembl_id"]


def get_chembl_metadata(query: str) -> Dict[str, Any]:
    """Retrieves metadata from ChEMBL for a SMILE string or ChEMBL ID.

    :param query: A valid SMILE string or ChEMBL ID, e.g.
        ``'CHEMBL1234'`` or ``'COc1cccc(c1)NC(=O)c2cccnc2'``
    :type query: str
    :return: A dictionary of metadata.
    :rtype: Dict[str, Any]
    """
    chembl = ChEMBL()
    return chembl.get_molecule(query)


def add_chembl_metadata(g: nx.Graph) -> nx.Graph:
    """
    Add ChEMBL metadata to the graph.

    :param g: The graph to add metadata to.
    :type g: nx.Graph
    :return: Graph with ChEMBL metadata added.
    :rtype: nx.Graph
    """
    # Get the SMILE
    smiles = g.graph["smiles"]
    g.graph["chembl_metadata"] = get_chembl_metadata(smiles)
    return g



"""Testing utilities for the Graphein library."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Any, Callable, Dict

import networkx as nx
import numpy as np
from loguru import logger as log

__all__ = [
    "compare_exact",
    "compare_approximate",
    "graphs_isomorphic",
    "nodes_equal",
    "edges_equal",
    "edge_data_equal",
]


def compare_exact(first: Dict[str, Any], second: Dict[str, Any]) -> bool:
    """Return whether two dicts of arrays are exactly equal.

    :param first: The first dictionary.
    :type first: Dict[str, Any]
    :param second: The second dictionary.
    :type second: Dict[str, Any]
    :return: ``True`` if the dictionaries are exactly equal, ``False``
        otherwise.
    :rtype: bool
    """
    if first.keys() != second.keys():
        return False
    return all(np.array_equal(first[key], second[key]) for key in first)


def compare_approximate(first, second):
    """Return whether two dicts of arrays are approximates equal.

    :param first: The first dictionary.
    :type first: Dict[str, Any]
    :param second: The second dictionary.
    :type second: Dict[str, Any]
    :return: ``True`` if the dictionaries are approx equal, ``False`` otherwise.
    :rtype: bool
    """
    if first.keys() != second.keys():
        return False
    return all(np.allclose(first[key], second[key]) for key in first)


def graphs_isomorphic(g: nx.Graph, h: nx.Graph) -> bool:
    """Checks for structural isomorphism between two graphs: ``g`` and ``h``.

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :return: ``True`` if the graphs are isomorphic, ``False`` otherwise.
    :rtype: bool
    """
    return nx.is_isomorphic(g, h)


def nodes_equal(g: nx.Graph, h: nx.Graph) -> bool:
    """Checks whether two graphs have the same nodes.

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :raises AssertionError: If the graphs do not contain the same nodes
    """
    for n in g.nodes():
        if n not in h.nodes():
            log.info(f"Node {n} (graph g) not in graph h")
            return False
    for n in h.nodes():
        if n not in g.nodes():
            log.info(f"Node {n} (graph h) not in graph g")
            return False
    return True


def edges_equal(g: nx.Graph, h: nx.Graph) -> bool:
    """Checks whether two graphs have the same edges.

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :raises AssertionError: If the graphs do not contain the same nodes
    """
    for u, v in g.edges():
        if (u, v) not in h.edges():
            log.info(f"Edge {u}-{v} (graph g) not in graph h")
            return False
    for u, v in h.edges():
        if (u, v) not in g.edges():
            log.info(f"Edge {u}-{v} (graph h) not in graph g")
            return False
    return True


def edge_data_equal(
    g: nx.Graph, h: nx.Graph, comparison_func: Callable = compare_exact
) -> bool:
    """Checks whether two graphs have the same edge features.

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :param comparison_func: Matching function for edge features.
        Takes two edge feature dictionaries and returns ``True`` if they are
        equal. Defaults to :func:`compare_exact`
    :type node_match_func: Callable
    :returns: ``True`` if the graphs have the same node features, ``False``
        otherwise.
    :rtype: bool
    """
    if not edges_equal(g, h):
        log.info("Edge lists do not match")
        return False
    for u, v in g.edges():
        if not compare_exact(g.edges[u, v], h.edges[u, v]):
            log.info(f"Edge {u}-{v} (graph g) features do not match graph h")
            return False
    for u, v in h.edges():
        if not compare_exact(g.edges[u, v], h.edges[u, v]):
            log.info(f"Edge {u}-{v} (graph h) features do not match graph g")
            return False
    return True


def node_data_equal(
    g: nx.Graph, h: nx.Graph, comparison_func: Callable = compare_exact
) -> bool:
    """Checks whether two graphs have the same node features.

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :param comparison_func: Matching function for node features.
        Takes two node dictionaries and returns True if they are equal.
        Defaults to :func:`compare_exact`
    :type comparison_func: Callable
    :returns: ``True`` if the graphs have the same node features, ``False``
        otherwise.
    :rtype: bool
    """
    if not nodes_equal(g, h):
        return False
    for n in g.nodes():
        if not compare_exact(g.nodes[n], h.nodes[n]):
            log.info(f"Node {n} (graph g) features do not match graph h")
            return False
    for n in h.nodes():
        if not compare_exact(g.nodes[n], h.nodes[n]):
            log.info(f"Node {n} (graph h) features do not match graph g")
            return False
    return True


def graphs_equal(
    g: nx.Graph,
    h: nx.Graph,
    node_match_func: Callable = compare_exact,
    edge_match_func: Callable = compare_exact,
) -> bool:
    """Asserts whether two graphs are equal
    (structural isomorphism and edge and node features match).

    :param g: The first graph.
    :type g: :class:`networkx.Graph`
    :param h: The second graph.
    :type h: :class:`networkx.Graph`
    :param node_match_func: Matching function for node features.
        Takes two node dictionaries and returns True if they are equal.
        Defaults to :func:`compare_exact`
    :type node_match_func: Callable
    :param edge_match_func: Matching function for edge features.
        A function that takes two edge dictionaries and returns ``True``
        if they are equal. Defaults to :func:`compare_exact`
    :type edge_match_func: Callable
    :return: ``True`` if the graphs are equal, ``False`` otherwise.
    :rtype: bool
    """
    return nx.is_isomorphic(g, h, node_match_func, edge_match_func)



"""Functions for working with Small Molecule Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import traceback
from functools import partial
from typing import Callable, Dict, List, Optional, Union

import networkx as nx
import numpy as np
from loguru import logger as log
from tqdm.contrib.concurrent import thread_map

from graphein.utils.dependencies import import_message, requires_python_libs
from graphein.utils.utils import (
    annotate_edge_metadata,
    annotate_graph_metadata,
    annotate_node_metadata,
    compute_edges,
)

from .chembl import get_smiles_from_chembl
from .config import MoleculeGraphConfig
from .edges.atomic import add_atom_bonds
from .utils import compute_fragments
from .zinc import get_smiles_from_zinc

try:
    import rdkit
    from rdkit import Chem
    from rdkit.Chem import AllChem
except ImportError:
    msg = import_message("graphein.molecule.graphs", "rdkit", "rdkit", True)
    log.warning(msg)


@requires_python_libs("rdkit")
def initialise_graph_with_metadata(
    name: str,
    rdmol: rdkit.Mol,
    coords: np.ndarray,
) -> nx.Graph:
    """
    Initializes the nx Graph object with initial metadata.

    :param name: Name of the molecule. Either the smiles or filename depending
        on how the graph was created.
    :type name: str
    :param rdmol: Processed DataFrame of molecule structure.
    :type rdmol: rdkit.Mol
    :return: Returns initial molecule structure graph with metadata.
    :rtype: nx.Graph
    """
    return nx.Graph(
        name=name, rdmol=rdmol, coords=coords, smiles=Chem.MolToSmiles(rdmol)
    )


@requires_python_libs("rdkit")
def add_nodes_to_graph(
    G: nx.Graph,
    verbose: bool = False,
) -> nx.Graph:
    """Add nodes into molecule graph.

    :param G: ``nx.Graph`` with metadata to populate with nodes.
    :type G: nx.Graph
    :param verbose: Controls verbosity of this step.
    :type verbose: bool
    :returns: nx.Graph with nodes added.
    :rtype: nx.Graph
    """
    for i, atom in enumerate(G.graph["rdmol"].GetAtoms()):
        coords = (
            G.graph["coords"][i] if G.graph["coords"] is not None else None
        )
        G.add_node(
            f"{atom.GetSymbol()}:{str(atom.GetIdx())}",
            atomic_num=atom.GetAtomicNum(),
            element=atom.GetSymbol(),
            rdmol_atom=atom,
            coords=coords,
        )

    if verbose:
        print(nx.info(G))
        print(G.nodes())

    return G


@requires_python_libs("rdkit")
def generate_3d(
    mol: Union[nx.Graph, Chem.Mol], recompute_graph: bool = False
) -> Union[nx.Graph, rdkit.Chem.rdchem.Mol]:
    """
    Generate a 3D structure for a RDKit molecule.

    Steps:
    1. Adds Hydrogens
    2. Embeds molecule with ``AllChem.ETKDGv3``
    (using ``useSmallRingTorsions=True``)
    3. Optimizes molecule with ``AllChem.MMFFOptimizeMolecule``
    4. Removes Hydrogens
    5. Returns molecule OR (optionally) recomputes the molecular graph
    (if ``recompute_graph=True``) using the new coordinates.

    :param mol: input molecule
    :type mol: Union[nx.Graph, Chem.Mol]
    :param recompute_graph: whether to recompute the graph based on the
        generated conformer.
    :type recompute_graph: bool
    :return: molecule with 3D coordinates or recomputed molecular graph.
    :rtype: Union[nx.Graph, Chem.Mol]
    """
    rdmol = mol.graph["rdmol"] if isinstance(mol, nx.Graph) else mol
    rdmol = Chem.AddHs(rdmol)
    params = AllChem.ETKDGv3()
    params.useSmallRingTorsions = True
    Chem.AddHs(rdmol)
    AllChem.EmbedMolecule(rdmol, params=params)
    AllChem.MMFFOptimizeMolecule(rdmol)
    rdmol = Chem.RemoveHs(rdmol)

    if recompute_graph:
        return construct_graph(config=mol.graph["config"], mol=rdmol)

    return rdmol


@requires_python_libs("rdkit")
def construct_graph(
    config: Optional[MoleculeGraphConfig] = None,
    mol: Optional[rdkit.Mol] = None,
    path: Optional[str] = None,
    smiles: Optional[str] = None,
    zinc_id: Optional[str] = None,
    chembl_id: Optional[str] = None,
    generate_conformer: Optional[bool] = False,
    edge_construction_funcs: Optional[str] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
) -> nx.Graph:
    """
    Constructs molecular structure graph from a ``sdf_path``, ``mol2_path``,
    ``smiles`` or RDKit Mol.

    Users can provide a :class:`~graphein.molecule.config.MoleculeGraphConfig`
    object to specify construction parameters.

    However, config parameters can be overridden by passing arguments directly
    to the function.

    :param config: :class:`~graphein.molecule.config.MoleculeGraphConfig`
        object. If None, defaults to config in ``graphein.molecule.config``.
    :type config: graphein.molecule.config.MoleculeGraphConfig, optional
    :param mol: rdkit.Mol object to build graph from. Defaults to ``None``.
    :type mol: rdkit.Mol, optional
    :param path: Path to either a ``.sdf``, ``.mol2``, ``.smi`` or ``pdb`` file. # noqa: E501
        Defaults to ``None``.
    :type path: str
    :param smiles: smiles string to build graph from. Default is ``None``.
    :type smiles: str, optional
    :param zinc_id: Zinc ID to build graph from. Default is ``None``.
    :type zinc_id: str, optional
    :param chembl_id: ChEMBL ID to build graph from. Default is ``None``.
    :type chembl_id: str, optional
    :param generate_conformer: Whether to generate a conformer for the molecule. # noqa: E501
        Defaults to ``False``.
    :type generate_conformer: bool, optional
    :param edge_construction_funcs: List of edge construction functions.
        Default is ``None``.
    :type edge_construction_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of edge annotation functions.
        Default is ``None``.
    :type edge_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of node annotation functions.
        Default is ``None``.
    :type node_annotation_funcs: List[Callable], optional
    :param graph_annotation_funcs: List of graph annotation function.
        Default is ``None``.
    :type graph_annotation_funcs: List[Callable]
    :return: Molecule Structure Graph
    :type: nx.Graph
    """

    # If no config is provided, use default
    if config is None:
        config = MoleculeGraphConfig()

    config.node_metadata_functions = (
        node_annotation_funcs
        if config.node_metadata_functions is None
        else config.node_metadata_functions
    )
    config.graph_metadata_functions = (
        graph_annotation_funcs
        if config.graph_metadata_functions is None
        else config.graph_metadata_functions
    )
    config.edge_metadata_functions = (
        edge_annotation_funcs
        if config.edge_metadata_functions is None
        else config.edge_metadata_functions
    )
    config.edge_construction_functions = (
        edge_construction_funcs
        if edge_construction_funcs is not None
        else config.edge_construction_functions
    )

    if zinc_id is not None:
        smiles = get_smiles_from_zinc(zinc_id)

    if chembl_id is not None:
        smiles = get_smiles_from_chembl(chembl_id)

    coords = None
    if smiles is not None:
        name = smiles
        rdmol = Chem.MolFromSmiles(smiles)
        if config.generate_conformer or generate_conformer:
            rdmol = generate_3d(mol=rdmol, recompute_graph=False)
            if config.add_hs:
                rdmol = Chem.AddHs(rdmol, addCoords=True)
            coords = get_coords(rdmol)

    if path is not None:
        name = path.split("/")[-1].split(".")[0]
        if path.lower().endswith(".sdf"):
            rdmol = Chem.SDMolSupplier(path)[0]
            if config.add_hs:
                rdmol = Chem.AddHs(rdmol, addCoords=True)
            coords = get_coords(rdmol)
        elif path.lower().endswith(".mol2"):
            rdmol = Chem.MolFromMol2File(path)
            if config.add_hs:
                rdmol = Chem.AddHs(rdmol, addCoords=True)
            coords = get_coords(rdmol)
        elif path.lower().endswith(".pdb"):
            name = path.split("/")[-1].split(".")[0]
            rdmol = Chem.MolFromPDBFile(path)
            if config.add_hs:
                rdmol = Chem.AddHs(rdmol, addCoords=True)
            coords = get_coords(rdmol)
        elif path.lower().endswith(".smi"):
            with open(path) as f:
                smiles = f.readlines()[0]
            name = smiles
            rdmol = Chem.MolFromSmiles(smiles)
            if config.generate_conformer or generate_conformer:
                rdmol = generate_3d(mol=rdmol, recompute_graph=False)
                if config.add_hs:
                    rdmol = Chem.AddHs(rdmol, addCoords=True)
                coords = get_coords(rdmol)

    elif mol is not None:
        name = Chem.MolToSmiles(mol)
        rdmol = mol
        if config.add_hs:
            rdmol = Chem.AddHs(rdmol, addCoords=True)
        coords = get_coords(rdmol)

    if coords is None:
        # If no coords are provided, add edges by bonds
        config.edge_construction_functions = [add_atom_bonds]
        g = initialise_graph_with_metadata(
            name=name,
            rdmol=rdmol,
            coords=None,
        )
    else:
        # If config params are provided, overwrite them
        config.edge_construction_functions = (
            edge_construction_funcs
            if config.edge_construction_functions is None
            else config.edge_construction_functions
        )

        g = initialise_graph_with_metadata(
            name=name,
            rdmol=rdmol,
            coords=np.asarray(coords),
        )

    # Add nodes to graph
    g = add_nodes_to_graph(g)

    # Add config to graph
    g.graph["config"] = config

    # Annotate additional node metadata
    if config.node_metadata_functions is not None:
        g = annotate_node_metadata(g, config.node_metadata_functions)

    # Compute graph edges
    g = compute_edges(
        g,
        funcs=config.edge_construction_functions,
    )

    # Annotate additional graph metadata
    if config.graph_metadata_functions is not None:
        g = annotate_graph_metadata(g, config.graph_metadata_functions)

    # Annotate additional edge metadata
    if config.edge_metadata_functions is not None:
        g = annotate_edge_metadata(g, config.edge_metadata_functions)

    return g


def compute_fragment_graphs(
    g: Union[nx.Graph, Chem.Mol], config: Optional[MoleculeGraphConfig] = None
) -> List[nx.Graph]:
    """Computes graphs for each fragment in a molecule.

    :param g: Input graph or molecule to fragment.
    :type g: Union[nx.Graph, Chem.Mol]
    :param config: Molecular graph construction config.
        See: :class:`graphein.molecule.config`. Defaults to ``None``.
    :type config: Optional[MoleculeGraphConfig], optional
    :return: List of fragment graphs.
    :rtype: List[nx.Graph]
    """
    if config is None:
        try:
            config = g.graph["config"]
        except KeyError:
            config = MoleculeGraphConfig()
    fragments = compute_fragments(g)
    return [construct_graph(mol=m, config=config) for m in fragments]


def _mp_graph_constructor(
    input: Union[str, Chem.Mol],
    config: MoleculeGraphConfig,
    use_mol: bool = False,
    use_file: bool = False,
    use_smiles: bool = False,
) -> nx.Graph:
    """
    Molecule graph constructor for use in multiprocessing several molecular
    graphs.

    :param args: Tuple of pdb code/path and the chain selection for that PDB
    :type args: Tuple[str, str]
    :param use_mol: Whether or not we are using RDKit Mols
    :type use_mol: bool
    :param use_file: Whether or not we are using file paths
    :type use_file: bool
    :param use_smiles: Whether or not we are using SMILES strings
    :type use_smiles: bool
    :param config: Molecule structure graph construction config
    :type config: MoleculeGraphConfig
    :return: Molecule structure graph
    :rtype: nx.Graph
    """
    func = partial(construct_graph, config=config)
    try:
        if use_mol:
            return func(mol=input)
        if use_file:
            return func(path=input)
        if use_smiles:
            return func(smiles=input)
    except Exception as e:
        log.info(
            f"Graph construction error for ({input})! {traceback.format_exc()}"
        )
        log.info(e)
        return None


def construct_graphs_mp(
    path_it: Optional[List[str]] = None,
    smiles_it: Optional[List[str]] = None,
    mol_it: Optional[List[Chem.Mol]] = None,
    config: MoleculeGraphConfig = MoleculeGraphConfig(),
    num_cores: int = 16,
    return_dict: bool = True,
) -> Union[List[nx.Graph], Dict[str, nx.Graph]]:
    """
    Constructs molecular graphs for a list of smiles or paths using
    multiprocessing.

    :param path_it: List of paths to use for molecule graph construction.
    :type path_it: Optional[List[str]], defaults to ``None``.
    :param smiles_it: List of smiles to use for molecule graph construction.
    :type smiles_it: Optional[List[str]], defaults to ``None``.
    :param mol_it: List of rdkit Mols to use for molecule graph construction.
    :type mol_it: Optional[List[Chem.Mol]], defaults to ``None``.
    :param config: MoleculeGraphConfig to use.
    :type config: graphein.molecule.config.MoleculeGraphConfig,
        defaults to default config params
    :param num_cores: Number of cores to use for multiprocessing.
        The more the merrier
    :type num_cores: int, defaults to 16
    :param return_dict: Whether or not to return a dictionary
        (indexed by smiles/paths) or a list of graphs.
    :type return_dict: bool, default to True
    :return: Iterable of molecule graphs. None values indicate there was a
        problem in constructing the graph for this particular molecule.
    :rtype: Union[List[nx.Graph], Dict[str, nx.Graph]]
    """
    assert (
        path_it is not None or smiles_it is not None or mol_it is not None
    ), "Iterable of paths, smiles strings OR RDKit mols is required."

    use_file, use_smiles, use_mol = False, False, False

    if path_it is not None:
        inputs = path_it
        use_file = True

    if smiles_it is not None:
        inputs = smiles_it
        use_smiles = True

    if mol_it is not None:
        inputs = mol_it
        use_mol = True

    constructor = partial(
        _mp_graph_constructor,
        use_file=use_file,
        use_mol=use_mol,
        use_smiles=use_smiles,
        config=config,
    )

    graphs = list(thread_map(constructor, list(inputs), max_workers=num_cores))

    if return_dict:
        graphs = {molecule: graphs[i] for i, molecule in enumerate(inputs)}

    return graphs


@requires_python_libs("rdkit")
def get_coords(mol: rdkit.Chem.rdchem.Mol) -> List[List[float]]:
    """
    Function for getting the coordinates of an RDKit molecule.

    :param mol: RDKit molecule.
    :type mol: rdkit.Chem.rdchem.Mol
    :return: Coordinates.
    :rtype: List[List[float]]
    """
    return [
        list(mol.GetConformer(0).GetAtomPosition(idx))
        for idx in range(mol.GetNumAtoms())
    ]
